#!/usr/bin/env python3
"""Unittest for module graph

In this modules, the new algorithm is tested. The TestCase GraphTesting checks
that the behavior of all important functions works as intended
"""
import unittest
from pathlib import Path
from typing import Any

import cobramod.core.graph as gr
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.debug import change_to_debug
from cobramod.error import GraphKeyError

# Debug must be set in level DEBUG for the test
change_to_debug()
# Setting directory for data
dir_data = Path(__file__).resolve().parent.joinpath("data")
# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class GraphTesting(unittest.TestCase):
    """
    This new TestCase Checks that the behavior of the new algorithm works as
    intended.
    """

    def test_find_missing(self):
        # CASE 1: Lineal
        test_dict = {"R1": "R2", "R2": "R3", "R3": None}
        gr.find_missing(graph=test_dict)
        # CASE 1b: Simple Lineal, Error
        test_dict = {"R1": ("R2", "R4"), "R2": "R3", "R3": None}
        self.assertRaises(GraphKeyError, gr.find_missing, graph=test_dict)

    def test_find_cycle(self):
        # CASE: Lineal
        test_dict = {"R1": "R2", "R2": "R3", "R3": None}
        test_answer = gr.find_cycle(graph=test_dict, key="R1", visited=[])
        self.assertListEqual(test_answer, [])

        # CASE: cyclic
        test_dict: dict[str, Any] = {"R1": "R2", "R2": "R3", "R3": "R1"}
        test_answer = gr.find_cycle(graph=test_dict, key="R1", visited=[])
        self.assertEqual(first=["R1", "R2", "R3"], second=test_answer)

        # CASE: Complex Lineal
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R5"),
            "R2": "R3",
            "R3": "R4",
            "R4": None,
            "R5": None,
        }
        test_answer = gr.find_cycle(graph=test_dict, key="R0", visited=[])
        self.assertEqual(test_answer, [])

        # CASE: Complex Cyclic
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R5"),
            "R2": "R3",
            "R3": "R4",
            "R4": "R0",
            "R5": None,
        }
        test_answer = gr.find_cycle(graph=test_dict, key="R0", visited=[])
        self.assertEqual(
            first=["R0", "R1", "R2", "R3", "R4"], second=test_answer
        )
        # CASE: Complex full Cycle
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R5"),
            "R2": "R3",
            "R3": "R4",
            "R4": "R0",
            "R5": "R4",
        }
        # Behavior of tuples
        test_answer = gr.find_cycle(graph=test_dict, key="R1", visited=[])
        self.assertEqual(first=len(test_answer), second=4)

    def test_return_cycles(self):
        # CASE: Lineal
        test_dict = {"R1": "R2", "R2": "R3", "R3": None}
        test_list = gr.return_cycles(graph=test_dict)
        self.assertFalse(expr=test_list)

        # CASE: Complex Cyclic
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R5"),
            "R2": "R3",
            "R3": "R4",
            "R4": "R0",
            "R5": None,
        }
        test_list = gr.return_cycles(graph=test_dict)
        self.assertIn(
            member=["R0", "R1", "R2", "R3", "R4"], container=test_list
        )

        # CASE: Graph with cycle, from Biocyc (GLUCONEO-PWY)
        test_dict = {
            "R1": "R13",
            "R2": "R3",
            "R3": ("R2", "R1"),
            "R4": None,
            "R5": "R4",
            "R6": "R5",
            "R7": "R6",
            "R8": "R7",
            "R9": "R8",
            "R10": None,
            "R11": "R10",
            "R12": None,
            "R13": None,
        }
        test_list = gr.return_cycles(graph=test_dict)
        self.assertIn(member=["R3", "R2"], container=test_list)
        # CASE: Complex Cycle (CALVIN-PWY)
        test_dict = {
            "R1": ("R11", "R2"),
            "R2": "R10",
            "R3": "R2",
            "R4": "R7",
            "R5": "R4",
            "R6": "R5",
            "R7": ("R8", "R13"),
            "R8": "R9",
            "R9": "R1",
            "R10": "R6",
            "R11": None,
            "R12": "R3",
            "R13": "R12",
        }
        test_list = gr.return_cycles(graph=test_dict)
        self.assertIn(
            ["R1", "R2", "R10", "R6", "R5", "R4", "R7", "R13", "R12", "R3"],
            test_list,
        )

    def test_cut_cycle(self):
        # CASE: Simple cut
        test_dict = {"R1": "R2", "R2": "R3", "R3": "R1"}
        gr.cut_cycle(graph=test_dict, key="R1")
        self.assertDictEqual(
            d1=test_dict, d2={"R1": None, "R2": "R3", "R3": "R1"}
        )

    def test_back(self):
        # CASE: Lineal
        test_dict = {"R1": "R2", "R2": "R3", "R3": None}
        test_answer = gr.back(graph=test_dict, value="R3", path=[])
        self.assertListEqual(list1=test_answer, list2=["R1", "R2"])

        # CASE: Complex Lineal
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R5"),
            "R2": "R3",
            "R3": "R4",
            "R4": None,
            "R5": None,
        }
        # For R5
        test_answer = gr.back(graph=test_dict, value="R5", path=[])
        self.assertListEqual(list1=test_answer, list2=["R0", "R1"])

        # For R4
        test_answer = gr.back(graph=test_dict, value="R4", path=[])
        self.assertListEqual(list1=test_answer, list2=["R0", "R1", "R2", "R3"])

        # CASE: Rest with stop list
        test_dict = {
            "R7": ("R8", "R11"),
            "R8": "R10",
            "R9": None,
            "R10": None,
            "R11": None,
        }
        test_answer = gr.back(
            graph=test_dict,
            value="R10",
            path=[],
            stop_list=["R0", "R1", "R2", "R3", "R4", "R5", "R6", "R12"],
        )
        self.assertListEqual(list1=test_answer, list2=["R7", "R8"])

        # CASE: Rest with stop list. Single element
        test_answer = gr.back(
            graph=test_dict,
            value="R9",
            path=[],
            stop_list=["R0", "R1", "R2", "R3", "R4", "R5", "R6", "R12"],
        )
        self.assertListEqual(list1=test_answer, list2=[])

    def test_verify_paths(self):
        # CASE: Complex Cycle (after modification)
        test_dict = {"R1": ("R11", "R2"), "R2": None, "R3": "R2", "R11": None}
        test_list = [["R1", "R2"]]
        test_answer = gr.verify_paths(paths=test_list, graph=test_dict)
        self.assertCountEqual(first=test_answer, second=["R3", "R11"])

    def test_get_paths(self):
        # CASE: Unrelated
        test_dict = {"R1": None, "R2": None, "R3": None}
        test_list = gr.get_paths(graph=test_dict, stop_list=[])
        self.assertEqual(first=len(test_list), second=3)
        self.assertIn(member=["R1"], container=test_list)
        self.assertIn(member=["R2"], container=test_list)
        self.assertIn(member=["R3"], container=test_list)

        # CASE: Simple Lineal
        test_dict = {"R1": "R2", "R2": "R3", "R3": None}
        test_list = gr.get_paths(graph=test_dict, stop_list=[])
        self.assertListEqual(
            list1=["R1", "R2", "R3"], list2=max(test_list, key=len)
        )
        # CASE: Complex Lineal
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R5"),
            "R2": "R3",
            "R3": "R4",
            "R4": None,
            "R5": None,
        }
        test_list = gr.get_paths(graph=test_dict, stop_list=[])
        self.assertListEqual(
            list1=["R0", "R1", "R2", "R3", "R4"], list2=max(test_list, key=len)
        )
        # CASE: Complex Lineal
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R7"),
            "R2": "R3",
            "R3": "R4",
            "R4": "R5",
            "R5": ("R6", "R9"),
            "R6": "R12",
            "R7": ("R8", "R11"),
            "R8": "R10",
            "R9": None,
            "R10": None,
            "R11": None,
            "R12": None,
        }
        test_list = gr.get_paths(graph=test_dict, stop_list=[])
        self.assertEqual(first=len(test_list), second=4)
        self.assertListEqual(
            list1=["R0", "R1", "R2", "R3", "R4", "R5", "R6", "R12"],
            list2=max(test_list, key=len),
        )
        # CASE: Using Rest
        test_dict = {
            "R7": ("R8", "R11"),
            "R8": "R10",
            "R9": None,
            "R10": None,
            "R11": None,
        }
        stop_list = (["R0", "R1", "R2", "R3", "R4", "R5", "R6", "R12"],)
        test_list = gr.get_paths(graph=test_dict, stop_list=stop_list)
        self.assertEqual(first=len(test_list), second=3)
        self.assertListEqual(
            list1=["R7", "R8", "R10"], list2=max(test_list, key=len)
        )
        # CASE: Complex Cycle (after modification)
        test_dict = {
            "R1": ("R11", "R2"),
            "R2": None,
            "R3": "R2",
            "R4": "R7",
            "R5": "R4",
            "R6": "R5",
            "R7": ("R8", "R13"),
            "R8": "R9",
            "R9": "R1",
            "R10": "R6",
            "R11": None,
            "R12": "R3",
            "R13": "R12",
        }
        test_list = gr.get_paths(graph=test_dict, stop_list=stop_list)
        # TODO: due to temporal fix, list is now 5 instead of 3
        self.assertEqual(first=len(test_list), second=5)

    def test_get_mapping(self):
        # CASE 0a: Single element
        test_dict = {"R1": None}
        test_list = gr.get_mapping(graph=test_dict, stop_list=[], new=[])
        self.assertListEqual(list1=test_list, list2=[["R1"]])
        # CASE 0b: Unrelated
        test_dict = {"R1": None, "R2": None, "R3": None}
        test_list = gr.get_mapping(graph=test_dict, stop_list=[], new=[])
        self.assertIn(member=["R1"], container=test_list)
        self.assertIn(member=["R2"], container=test_list)
        self.assertIn(member=["R3"], container=test_list)
        # CASE 1: Simple Lineal
        test_dict = {"R1": "R2", "R2": "R3", "R3": None}
        test_list = gr.get_mapping(graph=test_dict, stop_list=[], new=[])
        self.assertListEqual(list1=test_list, list2=[["R1", "R2", "R3"]])
        # CASE 2: Complex Lineal
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R5"),
            "R2": "R3",
            "R3": "R4",
            "R4": None,
            "R5": None,
        }
        test_list = gr.get_mapping(graph=test_dict, stop_list=[], new=[])
        self.assertListEqual(
            list1=test_list[0], list2=["R0", "R1", "R2", "R3", "R4"]
        )
        self.assertListEqual(list1=test_list[1], list2=["R5"])
        # CASE 2b: Complex Lineal
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R7"),
            "R2": "R3",
            "R3": "R4",
            "R4": "R5",
            "R5": ("R6", "R9"),
            "R6": "R12",
            "R7": ("R8", "R11"),
            "R8": "R10",
            "R9": None,
            "R10": None,
            "R11": None,
            "R12": None,
        }
        test_list = gr.get_mapping(graph=test_dict, stop_list=[], new=[])
        # Test the two longest maps
        self.assertListEqual(
            list1=test_list[0],
            list2=["R0", "R1", "R2", "R3", "R4", "R5", "R6", "R12"],
        )
        self.assertListEqual(list1=test_list[1], list2=["R7", "R8", "R10"])


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)



"""Mapping of graph

This modules handles the mapping of a directed graph into a matrix that can be
used in Escher. The main function is
:func:`cobramod.visualization.mapping.get_mapping`, which checks if the graph
is lineal and creates the representation in a matrix. In case of cyclic path,
it will cut it.
"""
from itertools import chain
from typing import Dict, List

from cobramod.core.graph import get_graph_dict


def get_all_values(dictionary: dict, keys: list) -> set:
    """
    Return a set with the all the values for a given list of keys. Elements of
    tuples will be added separately.

    Args:
        dictionary (dict): dictionary with keys and values
        keys (list): List of keys to get values

    Returns:
        Set: Values from given keys
    """
    set_values = set()
    for item in keys:
        value = dictionary[item]
        # In case of tuples
        if isinstance(value, tuple):
            for item in value:
                if item is not None:
                    set_values.add(item)
        else:
            if value is not None:
                set_values.add(value)
    return set_values


def child_map(mapping: list, dictionary: dict) -> dict:
    """
    Returns the relation parent: children as a dictionary. A key represent the
    parent and the key their corresponding children

    Args:
        mapping (list): Paths from given graph
        dictionary (dict): Original dictionary of directed graph.

    Returns:
        Dict: Relation of parent and children.
    """
    relation: Dict[str, list] = dict()
    for index, path in enumerate(mapping):
        # TODO: check if this condition is necesary
        # if len(path) == 1:
        #     continue
        # Check for each path
        path_2: list
        for index_2, path_2 in enumerate(mapping):
            if path_2 == path:
                continue
            # check if the first item is in the original path to check
            values = get_all_values(dictionary=dictionary, keys=path)
            if path_2[0] in values:
                try:
                    relation[str(index)].append(str(index_2))
                except KeyError:
                    relation[str(index)] = [str(index_2)]
    return relation


def get_index(dictionary: dict, path: list, value) -> int:
    """
    Return index of item in a path, whose value is found.

    Args:
        dictionary (dict): Original dictionary of directed graph
        path (list): A list with the name of nodes to be searched.
        value (Any): Value to find in path

    Returns:
        int: index of path.

    Raises:
        Warning: If value is not found.
    """
    for index, item in enumerate(path):
        values = get_all_values(dictionary=dictionary, keys=[item])
        if value in values:
            return index
    # TODO: add error
    raise Warning(f'Value "{value}" not found.')


def unformatted_matrix(graph: dict) -> List[list]:
    """
    Returns an unformatted matrix from a graph. The matrix represent the
    locations of the nodes and their relationships. Graph will be cut if
    cyclic. Unrelated nodes will be appended separately at the end. The matrix
    have 0 in empty positions.

    Args:
        graph (dict): Dictionary with relationship between nodes. A node can
            have multiple edges, which should be presented as values.
    Returns
        List[list]: Representation of matrix

    Raises:
        KeyError: If keys in graph are missing
    """
    # If graph cyclic, it will be modified
    mapping = get_graph_dict(graph=graph)
    if not mapping:
        return [[]]
    # By default firsts item
    longest = mapping[0]
    # TODO: change defaults of height
    relation = child_map(mapping=mapping, dictionary=graph)
    # If path is completely unrelated
    if not relation:
        return mapping
    matrix = [[0] * len(longest)]
    matrix[0] = longest
    # dictionary for start positions
    start_position = {"0": 0}
    for index, keys in relation.items():
        # child = item, parent = index
        item: str
        for item in keys:
            # Should not raised an error since we have the relation
            position_j = get_index(
                dictionary=graph,
                path=mapping[int(index)],
                # Always look for the first item
                value=mapping[int(item)][0],
            )
            # Use the start position from parent or define new start
            try:
                start = start_position[str(index)]
            except KeyError:
                start = 0
            # add 0's and extend
            start_position[item] = start + position_j + 1
            row = [0] * start_position[item]
            row.extend(mapping[int(item)])
            matrix.append(row)
    # Add unrelated paths
    for index, line in enumerate(mapping):
        if str(index) not in relation.keys() and str(index) not in list(
            chain.from_iterable(relation.values())
        ):
            matrix.append(line)
    return matrix


def fill_matrix(matrix: List[list], length: int):
    """
    Fills each row of the matrix with 0 until it reaches given length

    Args:
        matrix (List[list]): Matrix to fill
        length (int): Desired length.
    """
    for row in matrix:
        while len(row) < length:
            row.append(0)


def format_matrix(matrix: List[list], max_length: int) -> List[list]:
    """
    Formats given matrix and returns, if possible, a reduced matrix. Matrix
    will be filled with 0's if needed.

    Args:
        matrix (List[list]): Matrix to fill
        max_length (int): Desired length.
    """
    # TODO: check if sorting is necessary
    # Sort and fill missing 0
    matrix.sort(key=len, reverse=True)
    fill_matrix(matrix=matrix, length=max_length)
    # Squeeze rows with 0s if possible
    # new_matrix: List[list] = matrix.copy()
    for index_j, row in enumerate(matrix):
        # Find row with 0s
        if 0 not in row:
            # new_matrix.append(row)
            continue
        # Check if non-zero values can be appended above
        previous = {
            index_i
            for index_i, item in enumerate(matrix[index_j - 1])
            if item != 0
        }
        # check if all positions above are non-zero
        positions = {index_i for index_i, item in enumerate(row) if item != 0}
        # if position is in previous then there is no space
        if positions.issubset(previous):
            # new_matrix.append(row)
            continue
        previous_row: list = matrix[index_j - 1].copy()
        for index_i, item in enumerate(row):
            # Non-zero positions
            if index_i not in positions:
                continue
            previous_row[index_i] = row[index_i]
        # Remove previous and replace with new row
        matrix[index_j] = previous_row
        del matrix[index_j - 1]
    return matrix


def get_mapping(graph: dict) -> List[list]:
    """
    Returns a matrix for the representation of given graph.

    Args:
        graph (dict): Dictionary with relationship between nodes. A node can
            have multiple edges, which should be presented as values.
    Returns
        List[list]: Representation of matrix

    Raises:
        KeyError: If keys in graph are missing
    """
    # If graph is cyclic, it will be modified. Work with copy
    matrix = unformatted_matrix(graph=graph.copy())
    # Fill with 0 and merge rows if needed
    longest = len(matrix[0])
    matrix = format_matrix(matrix=matrix, max_length=longest)
    return matrix


def transpose(matrix: List[list]) -> List[list]:
    """
    Returns a transpose matrix for given matrix.
    """
    return list(map(list, zip(*matrix)))



"""Summary

This module is responsible for the short summary in the command
line and for other summaries in different formats.
"""
from __future__ import annotations
from pathlib import Path
from typing import Union, Optional

import pandas
from cobra import Model

from cobramod.debug import debug_log


class DataModel:
    """
    A class that can create a snapshot of a model and generates the summary
    based on those snapshots. Contains methods to identify differences and
    save them in multiple formats.
    """

    def __init__(self, lists: dict[str, list]):
        self.metabolites = lists.get("metabolites", [])
        self.demands = lists.get("demands", [])
        self.exchanges = lists.get("exchanges", [])
        self.genes = lists.get("genes", [])
        self.groups = lists.get("groups", [])
        self.sinks = lists.get("sinks", [])

        # reaction includes only values not already in exchanges or sinks

        self.reactions = [
            reaction
            for reaction in lists.get("reactions") or []
            if reaction not in set(self.sinks)
            and reaction not in set(self.exchanges)
        ]

    @classmethod
    def from_model(cls, model: Model):
        """
        Method to create a DataModel object based on a model object.

        Args:
            model (Model): Model based on which a DataModel object
                is to be created.
        """

        characteristics = {
            "reactions": model.reactions,
            "metabolites": model.metabolites,
            "demands": model.demands,
            "exchanges": model.exchanges,
            "genes": model.genes,
            "groups": model.groups,
            "sinks": model.sinks,
        }

        data = {}

        for identifier, values in characteristics.items():
            try:
                data[identifier] = values.list_attr("id")
            except AttributeError:
                data[identifier] = []

        return cls(data)

    def diff(self, other):
        """
        Creates a new DataModel object consisting of
        the differences between the original and the
        passed DataModel object.

        Args:
            other (DataModel): DataModel to be compared with this one.
        """
        left = self - other
        right = other - self
        data = {
            "reactions": left.reactions + right.reactions,
            "metabolites": left.metabolites + right.metabolites,
            "demands": left.demands + right.demands,
            "exchanges": left.exchanges + right.exchanges,
            "genes": left.genes + right.genes,
            "groups": left.groups + right.groups,
            "sinks": left.sinks + right.sinks,
        }

        return DataModel(data)

    def __sub__(self, other):
        reactions = set(other.reactions)
        metabolites = set(other.metabolites)
        demands = set(other.demands)
        exchanges = set(other.exchanges)
        genes = set(other.genes)
        groups = set(other.groups)
        sinks = set(other.sinks)

        reactions = [x for x in self.reactions if x not in reactions]
        metabolites = [x for x in self.metabolites if x not in metabolites]
        demands = [x for x in self.demands if x not in demands]
        exchanges = [x for x in self.exchanges if x not in exchanges]
        genes = [x for x in self.genes if x not in genes]
        groups = [x for x in self.groups if x not in groups]
        sinks = [x for x in self.sinks if x not in sinks]

        data = {
            "reactions": reactions,
            "metabolites": metabolites,
            "demands": demands,
            "exchanges": exchanges,
            "genes": genes,
            "groups": groups,
            "sinks": sinks,
        }

        return DataModel(lists=data)

    def __str__(self):
        output = [
            "Reactions:\n",
            str(self.reactions),
            "\n",
            "Metabolites:\n",
            str(self.metabolites),
            "\n",
            "Exchange:\n",
            str(self.exchanges),
            "\n",
            "Demand:\n",
            str(self.demands),
            "\n",
            "Sinks:\n",
            str(self.sinks),
            "\n",
            "Genes:\n",
            str(self.genes),
            "\n",
            "Groups:\n",
            str(self.groups),
            "\n",
        ]

        return "".join(output)

    def _to_dataframe(
        self, model: Model = None, additions=None, deletions=None
    ):
        """
        Creates a pandas DataFrame based on a DataModel object.
        You can pass another DataModel object to include it in
        the DataFrame to have both DataModels in one DataFrame.
        Intended for internal use only.

        Args:
            model (Model): Model for the extraction of model id and name.
            additions (DataModel): DataModel that contains the new entities
                in the model.
            deletions (DataModel): DataModel that contains the removed entities
                in the model.
        """
        dtype = pandas.StringDtype()

        dictionary = {
            "Reactions": pandas.Series(self.reactions, dtype=dtype),
            "Exchange": pandas.Series(self.exchanges, dtype=dtype),
            "Demand": pandas.Series(self.demands, dtype=dtype),
            "Sinks": pandas.Series(self.sinks, dtype=dtype),
            "Metabolites": pandas.Series(self.metabolites, dtype=dtype),
            "Genes": pandas.Series(self.genes, dtype=dtype),
            "Groups": pandas.Series(self.groups, dtype=dtype),
        }

        if model is not None:
            model_identifier = {
                "Model identifier": pandas.Series(str(model.id), dtype=dtype),
                "Model name": pandas.Series(str(model.name), dtype=dtype),
            }

            # Preserving the order so that the original model
            # comes first and then the modifications
            model_identifier.update(dictionary)
            dictionary = model_identifier

        if additions is not None:
            addition = {
                "New in Reactions": pandas.Series(
                    additions.reactions, dtype=dtype
                ),
                "New in Exchange": pandas.Series(
                    additions.exchanges, dtype=dtype
                ),
                "New in Demand": pandas.Series(additions.demands, dtype=dtype),
                "New in Sinks": pandas.Series(additions.sinks, dtype=dtype),
                "New in Metabolites": pandas.Series(
                    additions.metabolites, dtype=dtype
                ),
                "New in Genes": pandas.Series(additions.genes, dtype=dtype),
                "New in Groups": pandas.Series(additions.groups, dtype=dtype),
            }
            dictionary.update(addition)

        if deletions is not None:
            deleted = {
                "Removed in Reactions": pandas.Series(
                    additions.reactions, dtype=dtype
                ),
                "Removed in Exchange": pandas.Series(
                    additions.exchanges, dtype=dtype
                ),
                "Removed in Demand": pandas.Series(
                    additions.demands, dtype=dtype
                ),
                "Removed in Sinks": pandas.Series(additions.sinks, dtype=dtype),
                "Removed in Metabolites": pandas.Series(
                    additions.metabolites, dtype=dtype
                ),
                "Removed in Genes": pandas.Series(additions.genes, dtype=dtype),
                "Removed in Groups": pandas.Series(
                    additions.groups, dtype=dtype
                ),
            }
            dictionary.update(deleted)

        return pandas.DataFrame(dictionary)

    def to_excl(
        self, path, model: Model = None, additions=None, deletions=None
    ):
        """
        Method to save a DataModel as an Excel file.
        Can also be used to save the changes between
        two points in time, in an Excel file format.
        For other formats see :func:`cobramod.summary.DataModel.to_csv`
        or :func:`cobramod.summary.DataModel.to_txt`.

        Args:
            path (Path): Location where the file is to be saved.
            model (Model): Model for the extraction of model id and name.
            additions (DataModel): DataModel that contains the new entities in
                the model.
            deletions (DataModel): DataModel that contains the remote entities
                in the model.
        """

        save = self._to_dataframe(model, additions, deletions)
        save.to_excel(path, index=False)

    def to_csv(self, path, model: Model = None, additions=None, deletions=None):
        """
        Method to save a DataModel as a CSV file. Can also be used to save
        the changes between two points in time, as a CSV.
        For other formats see :func:`cobramod.summary.DataModel.to_excl` or
        :func:`cobramod.summary.DataModel.to_txt`.

        Args:
            path (Path): Location where the file is to be saved.
        model (Model): Model for the extraction of model id and name.
        additions (DataModel): DataModel that contains the new entities in
            the model.
        deletions (DataModel): DataModel that contains the remote entities
            in the model.
        """

        save = self._to_dataframe(model, additions, deletions)
        save.to_csv(path, index=False)

    def to_txt(self, path, model: Model = None, additions=None, deletions=None):
        """
        Method to save a DataModel as a txt file. Can also be used to save
        the changes between two points in time, as a txt.
        For other formats see :func:`cobramod.summary.DataModel.to_excl` or
        :func:`cobramod.summary.DataModel.to_csv`.

        Args:
            path (Path): Location where the file is to be saved.
            model (Model): Model for the extraction of model id and name.
            additions (DataModel): DataModel that contains the new entities
            in the model.
            deletions (DataModel): DataModel that contains the remote entities
            in the model.
        """
        output = []

        if model is not None:
            output.extend(
                [
                    "Summary:",
                    f"Model identifier: {model.id}",
                    "Model name:",
                    str(model.name),
                ]
            )

        output.append(str(self))

        if additions is not None:
            output.append("New:")
            output.append(str(additions))

        if deletions is not None:
            output.append("Removed:")
            output.append(str(deletions))

        with open(file=str(path), mode="w") as file:
            file.writelines(line + "\n" for line in output)


def summary(
    model: Model, original: DataModel, filename: Optional[Union[str, Path]]
):
    """
    Produces the short summary and another one in the defined format.

    Args:
        model (Model): model with recent changes.
        original (DataModel): Object with data from the previous model. Use
            method :func:`cobramod.summary.DataModel`.
        filename (Path): Location where the summary should be stored.
            The file format is determined by the suffix of the filename.
            Thus, '.txt', '.csv' or '.xlsx' can be used.
    """

    if isinstance(filename, str):
        filename = Path(filename)

    new_values = DataModel.from_model(model)

    deletions = original - new_values
    additions = new_values - original

    # check if there are any changes otherwise notify the user
    num_changes = 0
    for value in vars(additions).values():
        num_changes += len(value)

    for value in vars(deletions).values():
        num_changes += len(value)

    if num_changes == 0:
        msg = "No changes in the model were detected!"
        debug_log.warning(msg)

    # output a short summary
    print(
        "{:13} {:^7} | {:^7} {:10}".format(
            "Number of", "new", "removed", "entities in"
        )
        + "\n"
        + "*"
        + ("=" * 21)
        + "|"
        + ("=" * 19)
        + "*"
        + "\n" "{:13} {:^7} | {:^7} {:10}".format(
            "Reactions", len(additions.reactions), len(deletions.reactions), ""
        )
        + "\n" "{:13} {:^7} | {:^7} {:10}".format(
            "Metabolites",
            len(additions.metabolites),
            len(deletions.metabolites),
            "",
        )
        + "\n" "{:13} {:^7} | {:^7} {:10}".format(
            "Exchange", len(additions.exchanges), len(deletions.exchanges), ""
        )
        + "\n" "{:13} {:^7} | {:^7} {:10}".format(
            "Demand", len(additions.demands), len(deletions.demands), ""
        )
        + "\n" "{:13} {:^7} | {:^7} {:10}".format(
            "Sinks", len(additions.sinks), len(deletions.sinks), ""
        )
        + "\n" "{:13} {:^7} | {:^7} {:10}".format(
            "Genes", len(additions.genes), len(deletions.genes), ""
        )
        + "\n" "{:13} {:^7} | {:^7} {:10}".format(
            "Groups", len(additions.groups), len(deletions.groups), ""
        )
        + "\n"
    )

    # if desired, create another summary in the requested format
    if filename is None:
        return

    options = {
        ".xlsx": new_values.to_excl,
        ".csv": new_values.to_csv,
        ".txt": new_values.to_txt,
    }

    try:
        file_format = filename.suffix
        options[file_format](filename, model, additions, deletions)
    except KeyError:
        debug_log.warning(
            "Unknown format therefore no summary was created."
            "Use '.xlsx', ',csv' or ',txt'."
        )



"""Pathway extension

This module handles the addition of reactions as a pathway into
a model and the corresponding test that comes with it.

Most important functions:
- add_pathway: Adds a pathway or multiple reactions into a model.
- test_non_zero_flux: Checks that the given reaction in a model is active and
    gives a non-zero flux.
"""
from __future__ import annotations

from contextlib import suppress
from pathlib import Path
from typing import Generator, Optional, Union

import cobra.core as cobra_core
import cobra.exceptions as cobra_exceptions

import cobramod.retrieval as cmod_retrieval
import cobramod.utils as cmod_utils
from cobramod.core import creation as cmod_core_creation
from cobramod.core import graph as cmod_core_graph
from cobramod.core.pathway import Pathway
from cobramod.core.summary import DataModel
from cobramod.core.summary import summary as summarize
from cobramod.debug import debug_log

# Defaults to 1E-07
cobra_tolerance: float = cobra_core.Configuration().tolerance


def yield_reaction_from_list(
    sequence: list[str],
    compartment: str,
    directory: Path,
    database: str,
    show_imbalance: bool,
    stop_imbalance: bool,
    replacement: dict,
    model: cobra_core.Model,
    model_id: str,
    genome: Optional[str],
) -> Generator[cobra_core.Reaction, None, None]:
    """
    Yields a Reaction from given list of identifiers .

    .. hint:: Hyphens will become underscores. Double hyphens become single\
    underscores.

    Args:
        sequence (list): Identifiers for the reactions.
        directory (Path): Path to directory where data is located.
        database (str): Name of the database. Check
            :obj:`cobramod.available_databases` for a list of names.
        compartment (str): Location of the reaction.
        replacement (dict): Original identifiers to be replaced.
            Values are the new identifiers. This applies to metabolites.
        stop_imbalance (bool): If unbalanced reaction is found, stop process.
        show_imbalance (bool): If unbalanced reaction is found, show output.
        model_id (str): Exclusive for BIGG. Retrieve object from specified
            model.
        genome (str, optional): Exclusive for KEGG. Abbreviation for the
            specie involved. Genes will be obtained for this specie.

    Returns:
        Generator: New Reactions objects
    """

    debug_log.debug(f"Obtaining data for following reactions {sequence}.")
    # From given list (which could include None values), retrieve only Reaction
    # Objects.
    for identifier in sequence:
        try:
            obj = model.reactions.get_by_id(identifier)
        except KeyError:
            # Check if translation is available
            obj = cmod_core_creation.create_object(
                identifier=identifier,
                directory=directory,
                database=database,
                compartment=compartment,
                replacement=replacement,
                show_imbalance=show_imbalance,
                stop_imbalance=stop_imbalance,
                model=model,
                model_id=model_id,
                genome=genome,
            )
        if not isinstance(obj, cobra_core.Reaction):
            raise TypeError("Given object is not a valid COBRApy Reaction")
        yield obj


def find_problem(model: cobra_core.Model, identifier: str) -> list[str]:
    """
    Return a List with the metabolite identifiers that must have a sink
    reaction to make the reaction carry a non-zero flux.
    """
    problem: list[str] = list()
    value: Optional[float]

    reaction = model.reactions.get_by_id(identifier)
    if not isinstance(reaction, cobra_core.Reaction):
        raise TypeError("Given object is not a COBRApy Reaction")

    metabolites: list[cobra_core.Metabolite] = (
        reaction.products + reaction.reactants
    )

    # Add sink
    for metabolite in metabolites:
        with suppress(ValueError):
            model.add_boundary(metabolite=metabolite, type="sink")
            debug_log.debug(f"Sink reaction for {identifier} created.")

    assert model.slim_optimize() != 0.0

    # Test each reaction
    for metabolite in metabolites:
        sink = model.reactions.get_by_id(f"SK_{metabolite.id}")
        if not isinstance(sink, cobra_core.Reaction):
            raise TypeError("Given object is not a COBRApy Reaction")

        if len(metabolite.reactions) < 2:
            # Sink cannot be removed because it needs 2 reaction
            continue

        model.remove_reactions([sink])
        debug_log.debug(f"Sink reaction for {identifier} removed.")

        # Find metabolites that are necessary to carry a flux
        try:
            value = model.slim_optimize(error_value=None)

            if not value:
                value = 0.0

            if abs(value) <= cobra_tolerance:
                raise cobra_exceptions.OptimizationError("FAILED")

        except cobra_exceptions.OptimizationError:
            model.add_reactions(reaction_list=[sink])
            debug_log.debug(
                f"Sink reaction for {identifier} added again. Metabolite "
                f'"{metabolite.id}" must have a sink reaction to make '
                f'"{identifier}" carry a flux.'
            )
            assert metabolite.id
            problem.append(metabolite.id)

    return problem


def recursive_flux_test(
    model: cobra_core.Model, identifier: str, times: int = 0
) -> float:
    """
    Main part of the non-zero flux test. It is a recursive function. Firstly,
    the function checks if there a non-zero flux. If not, it will create the
    necessary sink reactions and rerun the test. This function warns the user
    if sink reactions were created.

    Args:
        model (Model): Model where the reactions are located
        reaction (str): identifier of the reaction
        times (int): Tracks how many times this function was called.

    Returns:
        Optional[float]: optimization value

    Raises:
        OptimizationError: Whenever a reaction needs manual intervention
    """

    if times == 0:
        debug_log.debug(
            f'Test to carry non-zero flux for "{identifier}" started.'
        )
    # run
    passed: bool
    value = model.slim_optimize()

    if not value:
        passed = False
        value = 0.0

    elif abs(value) > cobra_tolerance:
        passed = True

    else:
        passed = False

    # Add sinks and get new value
    if not passed:
        if times < 1:
            cmod_utils.confirm_sink(model=model, identifier=identifier)
            value = recursive_flux_test(
                model=model, identifier=identifier, times=times + 1
            )

        else:
            problem = find_problem(model=model, identifier=identifier)
            msg = (
                "The model cannot turnover the following metabolites "
                f"{problem}. To overcome this, sink reactions were created to "
                "simulate their synthesis."
            )
            debug_log.warning(msg)
            value = model.slim_optimize()

    # Raise only if manual intervention is necessary
    if times == 0:
        if abs(value) > cobra_tolerance:
            debug_log.info(
                f"Non-zero flux test for reaction '{identifier}' passed."
            )

        else:
            raise cobra_exceptions.OptimizationError(
                f"There is a problem with reaction {identifier}. Please create "
                "manually the corresponding turnover reactions for the "
                f'metabolites of "{identifier}".'
            )
    return value


def non_zero_core(model: cobra_core.Model, identifier: str):
    """
    Performs non-zero flux test. In this test, a reaction is tested to make
    sure it can carry a flux. If necessary, CobraMod creates sink reactions
    to simulate turnover of metabolites that participate in the reaction.
    Warnings will be shown for these cases.

    Args:
        model (Model): Model that contains the reaction
        reaction (str): identifier of the reaction

    Raises:
        OptimizationError: If given reaction needs manual curation

    """
    # Save old objective
    original_objective = model.objective
    original_direction = model.objective_direction

    reaction = model.reactions.get_by_id(identifier)

    # check reversibility
    if cmod_utils.reaction_is_minimize(model=model, identifier=identifier):
        model.objective_direction = "min"

    # Define new objective
    model.objective = reaction

    # Recursive function
    recursive_flux_test(model=model, identifier=identifier)

    # Revert back objective
    model.objective = original_objective
    model.objective_direction = original_direction


def test_non_zero_flux(model: cobra_core.Model, reaction: str):
    """
    Performs non-zero flux test. In this test, a reaction is tested to make
    sure it can carry a flux. If necessary, CobraMod creates sink reactions
    to simulate turnover of metabolites that participate in the reaction.
    Warnings will be shown for these cases.

    Args:
        model (Model): Model that contains the reaction
        reaction (str): identifier of the reaction

    Raises:
        OptimizationError: If given reaction needs manual curation

    """
    previous_sinks: set[str] = {sink.id for sink in model.sinks if sink.id}

    non_zero_core(model=model, identifier=reaction)

    # Revert back objective and inform
    cmod_utils.inform_new_sinks(model=model, previous_sinks=previous_sinks)
    print(f"Reaction {reaction} passed the non-zero-flux test.")


def add_reactions_to_Pathway(
    model: cobra_core.Model,
    pathway: Union[cobra_core.Group, Pathway],
    sequence: list[cobra_core.Reaction],
    ignore_list: list,
):
    """
    From a sequence of Reaction objects, add each Reaction into given model. It
    checks if new reactions do not break the optimized value. All reactions
    are added to a common CobraMod Pathway.

    Args:
        model (Model): Model to expand.
        pathway (Pathway): Common Pathway to add the reaction.
        sequence (list): List with :class:`cobra.core.reaction.Reaction`
            objects
        ignore_list (list, optional): A sequence of reactions that should be
            added but not tested for a non-zero-flux.

    Raises:
        TypeError: if reactions are not valid Reaction objects
    """
    if not all(
        (isinstance(reaction, cobra_core.Reaction) for reaction in sequence)
    ):
        raise TypeError("Reactions are not valid objects. Check list.")

    # Add sequence to model
    for reaction in sequence:
        # Add reaction if not in model
        if reaction not in model.reactions:
            model.add_reactions([reaction])
            debug_log.info(f'Reaction "{reaction.id}" was added to model.')

        # Skip test but include reaction in pathway
        if reaction.id not in ignore_list:
            non_zero_core(model=model, identifier=reaction.id)

        else:
            debug_log.warning(
                f'Reaction "{reaction.id}" found in "ignore_list". Skipping '
                + "non-zero flux test."
            )

        # Add to pathway only if reaction was not previously in the model.
        pathway.add_members(new_members=[reaction])
        debug_log.info(
            f'Reaction "{reaction.id}" added to group "{pathway.id}".'
        )

    debug_log.debug(f'Reactions added to group "{pathway.id}".')

    # Only add if there is at least 1 reaction in the group.
    if not model.groups.has_id(pathway.id) and len(pathway.members) > 0:
        model.add_groups(group_list=[pathway])
        debug_log.info(f'Pathway "{pathway.id}" added to Model.')


def remove_avoid_reactions(
    sequence: list[str], avoid_list: list[str]
) -> list[str]:
    """
    Updates the given sequence taking into consideration reactions to avoid
    in the sequence

    Args:
        sequence (list): identifier of the reactions to update
        avoid_list (list): Reactions that should not be included in the model
    """
    new_sequence: list[str] = list()

    # Remove reactions
    for item in sequence:
        if item in avoid_list:
            msg = (
                f'Reaction "{item}" was found in "avoid_list". Reaction {item}'
                + " will not be added to the model."
            )
            debug_log.warning(msg=msg)
            continue

        new_sequence.append(item)

    return new_sequence


def add_pathway_from_file(
    model: cobra_core.Model,
    file: Union[str, Path],
    identifier: str,
    database: Optional[str],
    replacement: dict,
    ignore_list: list,
    stop_imbalance: bool,
    show_imbalance: bool,
    directory: Path,
    genome: Optional[str],
    model_id: str,
):
    """
    Adds a Pathway into given model. The reactions are created from::

     - Path or str: A file with components. E. g. :
        Path.cwd().joinpath("file_with_names.txt") or "./file_with_names.txt"

      This applies for both options :

        :code:`reaction_identifier, compartment`

        For custom reactions

        :code:`reaction_identifier, reaction_name | coefficient metabolite <->
        coefficient metabolite

    Args:
        model (Model): Model to expand.
        path (Union[str, Path]): Location of the file.
        database (str): Name of the database.
            Check :obj:`cobramod.available_databases` for a list of names.
        identifier (str): Common :class:`cobramod.pathway.Pathway`
            identifier

    Argument for complex pathways:
        avoid_list (list, optional): A sequence of reactions identifiers to
            avoid adding to the model.
        replacement (dict, optional): Original identifiers to be replaced.
            Values are the new identifiers. This applies to metabolites as
            well. User can rename or replace identifiers using this argument.
        ignore_list (list, optional): A sequence of reactions that should be
            added but not tested for a non-zero-flux.

    Arguments for utilities:
        stop_imbalance (bool, optional): If an unbalanced reaction is found,
            stop the process. Defaults to False.
        show_imbalance (bool, optional): If an unbalanced reaction is found,
            print output. Defaults to True.
        model_id (str, optional): Exclusive for BIGG. Retrieve object from
            specified model. Pathways are not available.
            Defaults to: "universal"
        genome (str, optional): Exclusive for KEGG. Abbreviation for the
            species involved. Genes will be obtained for this species.
            List available at https://www.genome.jp/kegg/catalog/org_list.html

    """
    if isinstance(file, str):
        file = Path(file).absolute()

    # Either create a Pathway or give new name.
    try:
        pathway = model.groups.get_by_id(identifier)
        if not isinstance(pathway, cobra_core.Group) or not isinstance(
            pathway, Pathway
        ):
            raise TypeError(
                "Given object is not a valid COBRApy Group or a Cobramod Pathway"
            )
    except KeyError:
        pathway = Pathway(id=identifier)
        model.add_groups([pathway])

    # Get sinks
    previous_sinks: set[str] = {sink.id for sink in model.sinks if sink.id}

    # Get reactions from file
    new_reactions = cmod_core_creation.get_file_reactions(
        model=model,
        filename=file,
        directory=directory,
        database=database,
        replacement=replacement,
        stop_imbalance=stop_imbalance,
        show_imbalance=show_imbalance,
        genome=genome,
        model_id=model_id,
    )

    # Inform about sinks
    cmod_utils.inform_new_sinks(model=model, previous_sinks=previous_sinks)

    graph = cmod_core_graph.build_lineal_graph(
        sequence=[reaction.id for reaction in new_reactions]
    )

    add_reactions_to_Pathway(
        model=model,
        pathway=pathway,
        sequence=new_reactions,
        ignore_list=ignore_list,
    )

    # No need to format graph because there shouldn't be replacements
    if not pathway.graph:
        pathway.graph = graph
    else:
        pathway.graph.update(graph)
    pathway.notes["ORDER"] = pathway.graph


def add_pathway_from_data(
    model: cobra_core.Model,
    data: cmod_retrieval.Data,
    group: Optional[str],
    directory: Path,
    database: str,
    compartment: str,
    avoid_list: list,
    replacement: dict,
    ignore_list: list,
    stop_imbalance: bool,
    show_imbalance: bool,
    model_id: str,
    genome: Optional[str],
):
    """
    Adds a pathway into given model from a dictionary with the information of
    the pathway into given model from a dictionary with the information of
    the model.

    Args:
        model (Model): Model to expand.
        data_dict (dict): Dictionary with the information for the pathway.
        directory (Path): Path for directory to stored and retrieve data.
        database (str): Name of the database to search for reactions and
            metabolites. Check :obj:`cobramod.available_databases` for a list
            of names.
        compartment: Location of the reactions.

    Arguments for complex pathways:
        avoid_list (list, optional): A sequence of reactions identifiers to
            avoid adding to the model.
        replacement (dict, optional): Original identifiers to be replaced.
            Values are the new identifiers. This applies to metabolites as
            well. User can rename or replace identifiers using this argument.
        ignore_list (list, optional): A sequence of reactions that should be
            added but not tested for a non-zero-flux.

    Arguments for utilities:
        stop_imbalance (bool): If unbalanced reaction is found, stop process.
        show_imbalance (bool): If unbalanced reaction is found, show output.
        model_id (str, optional): Exclusive for BIGG. Retrieve object from
            specified model. Pathway are not available.
            Defaults to: "universal"
        genome (str, optional): Exclusive for KEGG. Abbreviation for the
            specie involved. Genes will be obtained from this specie.
            List available at https://www.genome.jp/kegg/catalog/org_list.html
    """
    # Create mapping from dictionary
    mapping = cmod_core_graph.get_graph_dict(data.attributes["pathway"])

    # Either create a Pathway or obtain the correct Pathway.
    identifier = data.identifier

    if group:
        identifier = group

    try:
        pathway = model.groups.get_by_id(identifier)
        if not isinstance(pathway, cobra_core.Group) or not isinstance(
            pathway, Pathway
        ):
            raise TypeError(
                "Given object is not a valid COBRApy Group or a Cobramod Pathway"
            )

    except KeyError:
        pathway = Pathway(id=identifier)
        model.add_groups([pathway])

    # Get sinks
    previous_sinks: set[str] = {sink.id for sink in model.sinks if sink.id}

    # FIXME: this should be changed
    graph: dict[str, Union[tuple[str], str, None]] = dict()

    for sequence in mapping:
        # Update sequence with removal of reactions
        sequence = remove_avoid_reactions(
            sequence=sequence, avoid_list=avoid_list
        )

        if not sequence:
            continue

        # Replacements will be used here
        sequence = list(
            yield_reaction_from_list(
                sequence=sequence,
                compartment=compartment,
                directory=directory,
                database=database,
                replacement=replacement,
                stop_imbalance=stop_imbalance,
                show_imbalance=show_imbalance,
                model=model,
                model_id=model_id,
                genome=genome,
            )
        )
        previous: str = ""
        for i, reaction in enumerate(sequence):
            if i == len(sequence) - 1:
                graph[reaction.id] = None

            if previous:
                value = graph.get(previous)

                if not value:
                    value = reaction.id

                elif isinstance(value, tuple):
                    a = list(value) + [reaction.id]
                    value = tuple(a)

                graph[previous] = value

            previous = reaction.id

        # Add to model
        add_reactions_to_Pathway(
            model=model,
            pathway=pathway,
            sequence=sequence,
            ignore_list=ignore_list,
        )

    # Inform about sinks
    cmod_utils.inform_new_sinks(model=model, previous_sinks=previous_sinks)

    if not pathway.graph:
        pathway.graph = graph

    else:
        pathway.graph.update(graph)
    pathway.notes["ORDER"] = pathway.graph


def add_pathway_from_strings(
    model: cobra_core.Model,
    identifier: str,
    sequence: list[str],
    database: str,
    compartment: str,
    directory: Path,
    avoid_list: list[str],
    replacement: dict[str, str],
    ignore_list: list[str],
    stop_imbalance: bool,
    show_imbalance: bool,
    model_id: str,
    genome: Optional[str],
):
    """
    Adds a sequence of identifiers to given model. It will automatically test
    for valid optimized values.

    Args:
        model (Model): Model to expand.
        identifier (str): Common :class:`cobramod.pathway.Pathway` identifier.
        sequence (list): List reaction identifiers.
        database (str): Name of the database.
            Check :obj:`cobramod.available_databases` for a list of names.
        compartment: Location of the reactions.
        directory (Path): Path for directory to store and retrieve data.

    Arguments for complex pathways:
        avoid_list (list, optional): A sequence of reactions identifiers to
            avoid adding to the model.
        replacement (dict, optional): Original identifiers to be replaced.
            Values are the new identifiers. This applies to metabolites as
            well. User can rename or replace identifiers using this argument.
        ignore_list (list, optional): A sequence of reactions that should be
            added but not tested for a non-zero-flux.

    Arguments for utilities:
        stop_imbalance (bool): If unbalanced reaction is found, stop process.
        show_imbalance (bool): If unbalanced reaction is found, show output.
        model_id (str, optional): Exclusive for BIGG. Retrieve object from
            specified model. Pathway are not available.
        genome (str): Exclusive for KEGG. Abbreviation for the
            specie involved. Genes will be obtained from this specie.
            List available at https://www.genome.jp/kegg/catalog/org_list.html
    """
    # Either create a Pathway or obtain the correct Pathway.
    try:
        pathway = model.groups.get_by_id(identifier)
        if not isinstance(pathway, cobra_core.Group):
            raise TypeError(
                f"The group for {identifier} is not a COBRApy Group object"
            )

    except KeyError:
        pathway = Pathway(id=identifier)
        model.add_groups([pathway])

    # Data storage is managed by the function
    sequence = remove_avoid_reactions(sequence=sequence, avoid_list=avoid_list)

    # Get sinks
    previous_sinks: set[str] = {sink.id for sink in model.sinks if sink.id}

    reactions = list(
        yield_reaction_from_list(
            sequence=sequence,
            compartment=compartment,
            database=database,
            replacement=replacement,
            directory=directory,
            stop_imbalance=stop_imbalance,
            show_imbalance=show_imbalance,
            model=model,
            model_id=model_id,
            genome=genome,
        )
    )
    # Create graph. It will be always simple lineal
    graph = cmod_core_graph.build_lineal_graph(
        sequence=[reaction.id for reaction in reactions]
    )

    # Append to model
    add_reactions_to_Pathway(
        model=model,
        sequence=reactions,
        pathway=pathway,
        ignore_list=ignore_list,
    )

    # Inform about sinks
    cmod_utils.inform_new_sinks(model=model, previous_sinks=previous_sinks)

    if isinstance(pathway, Pathway):
        if not pathway.graph:
            pathway.graph = graph
        else:
            pathway.graph.update(graph)
        pathway.notes["ORDER"] = pathway.graph


def add_pathway(
    model: cobra_core.Model,
    pathway: Union[list[str], str, Path],
    directory: Union[Path, str],
    compartment: str,
    database: Optional[str] = None,
    group: Optional[str] = None,
    avoid_list: list[str] = [],
    replacement: dict = {},
    ignore_list: list[str] = [],
    filename: Optional[Union[str, Path]] = None,
    stop_imbalance: bool = False,
    show_imbalance: bool = True,
    model_id: str = "",
    genome: Optional[str] = None,
):
    """
    Adds a pathway from the given database into a model. The argument 'pathway'
    can be a list of reactions identifiers or explicitly a pathway identifier.
    A group of reactions will be included in a custom group. The data will be
    downloaded and structured according to the database.

    Args:
        model (Model): Model to expand.
        pathway (list, str): Sequence of reaction identifiers or a single
            identifier for a pathway. Examples:
            ["RXN-2206", "RXN-207"], (Reaction to download from a database)
            "PWY-886", (Pathway to download from a database)
            ["ACALDt, RXN_2206_c"], (which are reactions already in the model)
        directory (Path): Path for the directory to store and retrieve data.
        database (str, Optional): Name of the database.
            Check :obj:`cobramod.available_databases` for a list of names.
            When adding reactions of the model, this argument is not necessary
            Defaults to None.
        compartment: Location of the reactions. If adding reaction already
            in the model, this argument will not change the reaction's
            compartment.
        group (str, optional): Common :class:`cobramod.pathway.Pathway`
            identifier. This will overwrite the name of the pathway.

    Arguments for complex pathways:
        avoid_list (list, optional): A sequence of reactions identifiers to
            avoid adding to the model.
        replacement (dict, optional): Original identifiers to be replaced.
            Values are the new identifiers. This applies to metabolites as
            well. User can rename or replace identifiers using this argument.
        ignore_list (list, optional): A sequence of reactions that should be
            added but not tested for a non-zero-flux.

    Arguments for summary:
        filename (Path, optional): Location for the summary. Defaults to
            "summary" in the current working directory. The file format is
            defined by the suffix. The suffixes '.txt', '.csv' and '.xlsx' can
            be used. If the filename is set to None, no summary will be
            created.

    Arguments for utilities:
        stop_imbalance (bool, optional): If an unbalanced reaction is found,
            stop the process. Defaults to False.
        show_imbalance (bool, optional): If an unbalanced reaction is found,
            print output. Defaults to True.
        model_id (str, optional): Exclusive for BIGG. Retrieve object from
            specified model. Pathways are not available.
            Defaults to: "universal"
        genome (str, optional): Exclusive for KEGG. Abbreviation for the
            species involved. Genes will be obtained for this species.
            List available at https://www.genome.jp/kegg/catalog/org_list.html
    """
    if not isinstance(model, cobra_core.Model):
        raise TypeError("Model is invalid")

    if isinstance(directory, str):
        directory = Path(directory).absolute()

    if not directory.exists():
        raise FileNotFoundError(
            f"Directory '{str(directory)}' not found. Create the data directory"
        )

    # Save information for summary methods
    old_values = DataModel.from_model(model)

    # Check if identifier
    if isinstance(pathway, str):
        identifier = pathway

        pathway = Path(pathway).absolute()

        # Identifier found
        if not pathway.suffix:
            pathway = identifier

    if isinstance(pathway, str):
        data_dict = cmod_retrieval.get_data(
            directory=directory,
            identifier=pathway,
            database=database,
            model_id=model_id,
            genome=genome,
        )
        if not database:
            raise AttributeError(
                "Database argument cannot be empty. Specify the name"
            )
        add_pathway_from_data(
            model=model,
            data=data_dict,
            directory=directory,
            database=database,
            compartment=compartment,
            avoid_list=avoid_list,
            replacement=replacement,
            ignore_list=ignore_list,
            show_imbalance=show_imbalance,
            stop_imbalance=stop_imbalance,
            model_id=model_id,
            genome=genome,
            group=group,
        )

    elif isinstance(pathway, Path):
        if not group:
            group = "custom_group"

        add_pathway_from_file(
            model=model,
            file=pathway,
            database=database,
            ignore_list=ignore_list,
            genome=genome,
            model_id=model_id,
            directory=directory,
            replacement=replacement,
            show_imbalance=show_imbalance,
            stop_imbalance=stop_imbalance,
            identifier=group,
        )

    elif isinstance(pathway, list):
        if not group:
            group = "custom_group"

        # if not database:
        #     raise AttributeError(
        #         "Database argument cannot be empty. Specify the name"
        #     )

        add_pathway_from_strings(
            model=model,
            identifier=group,
            sequence=pathway,
            compartment=compartment,
            avoid_list=avoid_list,
            directory=directory,
            database=database,
            replacement=replacement,
            ignore_list=ignore_list,
            show_imbalance=show_imbalance,
            stop_imbalance=stop_imbalance,
            model_id=model_id,
            genome=genome,
        )
    else:
        raise ValueError("Argument 'pathway' must be iterable or a identifier")

    summarize(model, old_values, filename=filename)



"""Object creation

This module handles the creation of COBRApy's objects
:class:`cobra.core.metabolite.Metabolite` and
:class:`cobra.core.reaction.Reaction`. Dictionaries containing the data of a
given database are used.
"""
from __future__ import annotations

from contextlib import suppress
from pathlib import Path
from typing import Optional, Union

import cobra.core as cobra_core
import requests

import cobramod.error as cmod_error
import cobramod.retrieval as cmod_retrieval
import cobramod.utils as cmod_utils
from cobramod.core import crossreferences
from cobramod.debug import debug_log


def build_metabolite(
    identifier: str, formula: str, name: str, charge: float, compartment: str
) -> cobra_core.Metabolite:
    """
    Returns a basic :class:`cobra.Metabolite`.

    LOG level: Debug

    Args:
        identifier (str): Short name for Metabolite
        formula (str): Chemical formula
        name (name): Long name for Metabolite
        charge (float): Charge of the metabolite
        compartment (str): Location

    Returns:
        Metabolite
    """
    metabolite = cobra_core.Metabolite(
        id=identifier,
        formula=formula,
        name=name,
        charge=charge,
        compartment=compartment,
    )
    debug_log.debug(f'Manually curated metabolite "{identifier}" was created.')
    return metabolite


def metabolite_from_string(
    line_string: str, replacement: dict[str, str], model: cobra_core.Model
) -> cobra_core.Metabolite:
    """
    Creates a Metabolite object based on a string. Function will try to find
    the metabolite in the model or cross-reference and return it if possible.

    The string must follow the syntax:

    :code:`formatted identifier, name , compartment, chemical_formula,
    molecular_charge`

    Args:
        line_string (str): string with information
        replacement (dict[str, str]): Dictionary with either the new identifier
            and/or the identifier of an object inside the model
        model (Model): Model to search for cross-references

    Raises:
        WrongSyntax: if format is does not follow syntax

    Returns:
        Metabolite
    """
    line = [part.strip().rstrip() for part in line_string.split(",")]

    if len(line) != 5:
        raise cmod_error.WrongSyntax(
            "Given line is invalid. Format is:\nid, name, compartment, "
            f"chemical_formula, molecular_charge.\nVerify line:\n{line_string}"
        )

    identifier = line[0]

    # Try to find replacement
    new_identifier = find_replacements(
        identifier=identifier,
        obj_type="metabolites",
        replace_dict=replacement,
        model=model,
    )
    if isinstance(new_identifier, cobra_core.Metabolite):
        return new_identifier

    with suppress(KeyError):
        identifier = replacement[identifier]

    name = line[1]
    compartment = line[2]
    formula = line[3]
    charge = line[4]

    debug_log.debug(
        f'Metabolite "{identifier}" was identified as a manually. '
        + "curated metabolite."
    )
    return build_metabolite(
        identifier=identifier,
        name=name,
        compartment=compartment,
        charge=float(charge),
        formula=formula,
    )


def metabolite_from_data(
    data: cmod_retrieval.Data,
    compartment: str,
    model: cobra_core.Model = cobra_core.Model(),
) -> cobra_core.Metabolite:
    """
    Parses the metabolite information from a Data object and creates a
    Metabolite in given compartment. This function will try to figure out if
    the metabolite is already in the model

    Returns:
        Metabolite
    """
    # Try to obtain available translation
    xref = cmod_utils.find_intersection(
        dictlist=model.metabolites,
        query=data.attributes["xref"],
        revert=True,
    )

    if xref:
        data.identifier = xref

    # No need for replacement dictionary for metabolites
    metabolite = data.parse(model, compartment)
    if not isinstance(metabolite, cobra_core.Metabolite):
        raise AttributeError("Given object is not a Metabolite")

    debug_log.debug(f"Metabolite '{metabolite.id}' created from Data object")
    return metabolite


def convert_string_metabolite(
    line: str,
    model: cobra_core.Model,
    replacement: dict[str, str],
    directory: Path,
    database: Optional[str],
    model_id: Optional[str],
) -> cobra_core.Metabolite:
    """
    Transform a string into a Metabolite and returns it. This object can be
    either manually-curated or from a database. If the compound if found under
    a different name (cross-reference), this will be returned instead.

    Custom metabolite syntax:

    :code:`formatted_identifier, name, compartment, chemical_formula,
    molecular_charge`

    Metabolite from database:

    :code:`metabolite_identifier, compartment`

    Args:
        line (str): string with information of metabolite
        model (Model): Model to check for cross-references
        replacement (dict[str, str]): Dictionary with either the new identifier
            and/or the identifier of an object inside the model
        directory (Path): Path to directory where data is located
        database (str, optional): Name of database. Check
            :obj:`cobramod.retrieval.available_databases` for a list of names.
            The argument can be None ONLY for manually curated metabolites
        model_id (str, optional): Exclusive for BIGG. Name of model

    Returns:
        Metabolite
    """
    if line.count(",") > 1:
        new_metabolite = metabolite_from_string(
            line_string=line, replacement=replacement, model=model
        )
    else:
        # Prep string
        segment = (part.strip().rstrip() for part in line.split(sep=","))
        identifier = next(segment)
        compartment = next(segment)

        if replacement:
            old_metabolite = find_replacements(
                identifier=identifier,
                replace_dict=replacement,
                model=model,
                obj_type="metabolites",
            )
            if isinstance(old_metabolite, cobra_core.Metabolite):
                return old_metabolite

        identifier = replacement.get(identifier, identifier)

        # Retrieve from database
        data = cmod_retrieval.get_data(
            identifier, directory, database, model_id
        )
        new_metabolite = data.parse(model, compartment, replacement)

    if not isinstance(new_metabolite, cobra_core.Metabolite):
        raise TypeError("Given object is not a valid COBRApy Metabolite")
    return new_metabolite


def get_file_metabolites(
    model: cobra_core.Model,
    filename: Path,
    replacement: dict[str, str],
    directory: Path,
    database: Optional[str],
    model_id: Optional[str],
) -> list[cobra_core.Metabolite]:
    """
    Return a list with Metabolites from a file. If a metabolite is found in
    model under a different name (cross-reference), it will be included in the
    list instead.

    Custom metabolite syntax:

    :code:`formatted_identifier, name, compartment, chemical_formula,
    molecular_charge`

    Metabolite from database:

    :code:`metabolite_identifier, compartment`

    Args:
        model (Model): Model to get metabolites if available
        filename (Path): location of the file with metabolites
        replacement (dict[str, str]): Dictionary with either the new identifier
            and/or the identifier of object inside the model.
        directory (Path): Path to directory where data is located.
        database (str, optional): Name of database. Check
            :obj:`cobramod.retrieval.available_databases` for a list of names.
            The argument can be None ONLY for manually curated metabolites
        model_id (str, optional): Bigg-specific argument. Name of the model

    Raises:
        FileNotFoundError: If given file is not found

    Returns:
        list[Metabolite]
    """
    if not filename.exists():
        raise FileNotFoundError(
            f'Given file in "{str(filename)}" does not exists. '
            + "Please create the given file or verify that the path is correct"
        )
    # For each line, build and add metabolite. If a Metabolite is no properly
    # created, either raise an Error or use a default.
    with open(filename, "r") as f:
        new_metabolites: list[cobra_core.Metabolite] = list()
        for line in cmod_utils.read_lines(f=f):
            new_metabolites.append(
                convert_string_metabolite(
                    line, model, replacement, directory, database, model_id
                )
            )
    return new_metabolites


def get_reaction(
    data: cmod_retrieval.Data,
    compartment: str,
    replacement: dict[str, str],
    show_imbalance: bool,
    stop_imbalance: bool,
    model: cobra_core.Model,
) -> cobra_core.Reaction:
    """
    Creates a :class:`cobra:Reaction` from reaction Data. Location of the
    reaction can be set with the argument 'compartment'.

    This function searchs if the reaction and their metabolites are already in
    the model and will use them instead. Additionally, genes will be created
    and linked to the reaction.

    Args:
        data_dict (Data): Biochemical information of the reaction
        compartment (str): Locations of the reaction
        replacement (dict[str, str]): Original identifiers to be replaced
            Values are the new identifiers.
        model (Model): Model to search for duplicates and cross-references
        stop_imbalance (bool): If unbalanced reaction is found, stop process
        show_imbalance (bool): If unbalanced reaction is found, show output

    Returns:
        Reaction
    """
    if data.database == "META":
        msg = (
            f'Metabolic pathway information for reaction "{data.identifier}" '
            'comes from database "META". No genes will be created. Please use '
            "a specie-specific sub-database from BioCyc."
        )
        debug_log.warning(msg)

    xref = cmod_utils.find_intersection(
        dictlist=model.reactions,
        query=data.attributes["xref"],
        revert=True,
    )

    if xref:
        data.identifier = xref

    data.attributes["replacements"] = replacement
    reaction = data.parse(model, compartment, replacement)

    if not isinstance(reaction, cobra_core.Reaction):
        raise AttributeError("Given object is not a Metabolite")

    cmod_utils.check_imbalance(
        reaction=reaction,
        show_imbalance=show_imbalance,
        stop_imbalance=stop_imbalance,
    )

    return reaction


def reaction_from_string(
    line_string: str,
    directory: Path,
    database: Optional[str],
    stop_imbalance: bool,
    show_imbalance: bool,
    replacement: dict[str, str],
    model_id: Optional[str],
    model: cobra_core.Model = cobra_core.Model(),
) -> cobra_core.Reaction:
    """
    Returns a custom reaction from a string that includes the information
    of the reaction and its metabolites.

    If the metabolites are not in inside the model, they will be retrieved from
    the specified database. This Function will also search for cross-references
    in the model.

    Syntax:

    :code:`reaction_identifier, reaction_name | coefficient metabolite <->
    coefficient metabolite

    Identifiers of metabolites have to end with an underscore and a
    compartment:

    E.g **`4 OXYGEN-MOLECULE_c`**

    Else, include a model to retrieve identifiers from it.

    Reversibility will depend on the type of arrow. Options are:

    **`<--, -->, <=>, <->`**

    Args:
        line_string (str): string with information
        directory (Path): Path to directory where data is located
        database (str, optional): Name of database. Check
            :obj:`cobramod.retrieval.available_databases` for a list of names.
            The argument can be None ONLY for manually curated metabolites
        stop_imbalance (bool): If unbalanced reaction is found, stop process
        show_imbalance (bool): If unbalanced reaction is found, show output
        replacement (dict): Dictionary with either the new identifier and/or
            the identifier of an object inside the model
        model_id (optional, str): Bigg-specific argument. Name of the model
        model (Model, optional): A model to obtain metabolite objects from.
            Defaults to an empty Model

    Raises:
        NoDelimiter: If no delimiter "|" is found.

    Returns:
        Reaction
    """
    # Remove blank and obtains information
    try:
        info, reaction_str = [
            string.strip().rstrip() for string in line_string.split("|")
        ]
    except ValueError:
        raise cmod_error.NoDelimiter(line_string)

    # It is possible to pass only the identifier
    if "," in info:
        reaction_id, reaction_name = [
            string.strip().rstrip() for string in info.split(",")
        ]
    else:
        reaction_id = reaction_name = info

    # Create Base reaction and then fill it with its components.
    reaction = cobra_core.Reaction(id=reaction_id, name=reaction_name)

    cmod_retrieval.build_reaction_from_str(
        model,
        reaction,
        reaction_str,
        directory,
        database,
        model_id,
        replacement,
    )

    # Set new bounds and check imbalance
    # new_reaction.bounds = bounds
    cmod_utils.check_imbalance(
        reaction=reaction,
        show_imbalance=show_imbalance,
        stop_imbalance=stop_imbalance,
    )
    return reaction


def string_to_reaction(
    line: str,
    model: cobra_core.Model,
    directory: Path,
    database: Optional[str],
    stop_imbalance: bool,
    show_imbalance: bool,
    replacement: dict[str, str],
    model_id: Optional[str],
    genome: Optional[str],
) -> cobra_core.Reaction:
    """
    Returns a :class:`cobra.Reaction` from a string. It can be either
    manually-curated, or the identifier for a reaction in a database. The
    function will search for cross-references of the reaction and its
    metabolites if the database provides that information

    Syntax:

    :code:`reaction_identifier, compartment`

    For custom reactions:

    :code:`reaction_identifier, reaction_name | coefficient metabolite <->
    coefficient metabolite

    Identifiers of metabolites have to end with an underscore and a
    compartment:

    E.g **`4 OXYGEN-MOLECULE_c`**

    Else, include a model to retrieve identifiers from it.

    Reversibility will depend on the type of arrow that. Options are

    **`<--, -->, <=>, <->`**

    Args:
        line (str): String with custom reaction or identifier of reaction
        model (Model): Model to search for cross-references
        directory (Path): Path to directory, where data is located
        database (str, optional): Name of database. Check
            :obj:`cobramod.retrieval.available_databases` for a list of names.
            This argument can be empty ONLY for manually-curated reactions
        stop_imbalance (bool): If unbalanced reaction is found, stop process
        show_imbalance (bool): If unbalanced reaction is found, show output
        replacement (dict[str, str]): Dictionary with either the new identifier
            and/or the identifier of an object inside the model.
        genome (str, optional): Exclusive for KEGG. Abbreviation for the
            specie involved. Genes will be obtained from this specie
        model_id (str, optional): Exclusive for BIGG. Retrieve object from
            specified model

    Returns:
        Reaction
    """
    if "|" in line:
        # Create custom reaction
        reaction = reaction_from_string(
            line_string=line,
            directory=directory,
            database=database,
            model=model,
            stop_imbalance=stop_imbalance,
            show_imbalance=show_imbalance,
            replacement=replacement,
            model_id=model_id,
        )

    else:
        # Must obtain from database
        identifier, compartment = (
            part.strip().rstrip() for part in line.split(",")
        )

        if database is None:
            raise TypeError(
                "Function 'string_to_reaction' cannot retrieve metabolic "
                "pathway information if no database if provided"
            )

        reaction = create_object(
            identifier,
            directory,
            database,
            compartment,
            replacement,
            show_imbalance,
            stop_imbalance,
            model,
            model_id,
            genome,
        )

        if not isinstance(reaction, cobra_core.Reaction):
            raise AttributeError("Given object is not a valid COBRApy Reaction")

        cmod_utils.check_imbalance(reaction, stop_imbalance, show_imbalance)

    return reaction


def get_file_reactions(
    model: cobra_core.Model,
    filename: Union[str, Path],
    directory: Path,
    stop_imbalance: bool,
    show_imbalance: bool,
    replacement: dict[str, str],
    database: Optional[str],
    model_id: Optional[str],
    genome: Optional[str],
) -> list[cobra_core.Reaction]:
    """
    Returns list with reactions from file. All reactions can be either created
    manually or retrieved from a database. For each reactions, its always
    checks for mass balance.

    Custom reactions:

    :code:`reaction_identifier, reaction_name | coefficient metabolite <->
    coefficient metabolite

    Identifiers of metabolites have to end with an underscore and a
    compartment:

    E.g **`4 OXYGEN-MOLECULE_c`**

    Else, include a model to retrieve identifiers from it. Reversibility will
    depend on the type of arrow that. Options are:

    **`<--, -->, <=>, <->`**

    From database:

    :code:`original_identifier, compartment`


    Args:
        model (Model): model to check for cross-references
        filename (Path, str): location of the file with reaction information
        directory (Path): Path to directory where data is located
        stop_imbalance (bool): If unbalanced reaction is found, stop process
        show_imbalance (bool): If unbalanced reaction is found, show output
        replacement (dict[str, str]): Dictionary with either the new identifier
            and/or the identifier of an object inside the model
        database (str, optional): Name of database. Check
            :obj:`cobramod.retrieval.available_databases` for a list of names.
            This argument can be empty ONLY for manually-curated reactions
        model_id (str, optional): Exclusive for BIGG. Retrieve object from
            specified model
        genome (str, optional): Exclusive for KEGG. Abbreviation for the
            specie involved. Genes will be obtained from this specie

    Raises:
        FileNotFoundError: If file does not exists

    Returns:
        list[Reaction]
    """
    if isinstance(filename, str):
        filename = Path(filename).absolute()

    if not filename.exists():
        raise FileNotFoundError(
            f'File "{filename.name}" does not exist. '
            + "Check if the given path is correct."
        )
    with open(filename, "r") as f:
        new_reactions = list()
        for line in cmod_utils.read_lines(f=f):
            new_reactions.append(
                string_to_reaction(
                    line=line,
                    model=model,
                    directory=directory,
                    database=database,
                    stop_imbalance=stop_imbalance,
                    show_imbalance=show_imbalance,
                    replacement=replacement,
                    model_id=model_id,
                    genome=genome,
                )
            )
    return new_reactions


def create_object(
    identifier: str,
    directory: Union[Path, str],
    database: str,
    compartment: str,
    replacement: dict[str, str] = {},
    show_imbalance: bool = True,
    stop_imbalance: bool = False,
    model: cobra_core.Model = cobra_core.Model(),
    model_id: Optional[str] = None,
    genome: Optional[str] = None,
) -> cobra_core.Object:
    """
    Creates and returns a COBRApy object based on given identifier and
    database. Identifier names will be formatted.

    .. hint:: Hyphens will become underscores.

    Args:
        identifier (str): Original identifier of the database.
        directory (Path): Path to directory where data is stored.
        database (str): Name of database. Check
            :obj:`cobramod.retrieval.available_databases` for a list of names.
        compartment (str): Location of the object. In case of reaction, all
            metabolites will be included in the same location.
        replacement (dict[str, str]): Dictionary with either the new identifier
            and/or the identifier of an object inside the model.

    Arguments for reactions:
        stop_imbalance (bool): If an unbalanced reaction is found, stop the
            process. Defaults to False.
        show_imbalance (bool): If an unbalanced reaction is found, print
            output. Defaults to True.
        model (Model): Model in which cross-references for reaction
            and metabolites might be found and used. If nothing is specified,
            it uses an empty model by default

    Special arguments for databases:
        model_id (str, optional): Exclusive for BIGG. Retrieve object from
            specified model. Internal functions use the default: "universal"
        genome (str, optional): Exclusive for KEGG. Abbreviation for the
            species involved. Genes will be obtained for this species.
            List available at https://www.genome.jp/kegg/catalog/org_list.html

    Returns:
        cobra.core.Object: A Reaction or Metabolite object
    """
    query = replacement.get(identifier, identifier)

    if not isinstance(directory, Path):
        directory = Path(directory).absolute()

    try:
        data = cmod_retrieval.get_data(
            directory=directory,
            identifier=query,
            database=database,
            # debug_level=10,
            model_id=model_id,
            genome=genome,
        )
    except requests.HTTPError:
        # If not found, then rename
        data = cmod_retrieval.get_data(
            directory=directory,
            identifier=identifier,
            database=database,
            # debug_level=10,
            model_id=model_id,
            genome=genome,
        )
        data.identifier = replacement.get(identifier, identifier)

    if data.mode == "Reaction":
        obj = get_reaction(
            data,
            compartment,
            replacement,
            show_imbalance,
            stop_imbalance,
            model,
        )
    else:
        obj = data.parse(model, compartment, replacement)

    return obj


def add_metabolites(
    model: cobra_core.Model,
    obj: Union[Path, str, list[cobra_core.Object], list[str]],
    directory: Union[Path, str] = Path.cwd().joinpath("data"),
    database: Optional[str] = None,
    **kwargs,
):
    """
    Adds given object into the model. The options are:

     - Path or str: A file with the metabolites. E. g:
        Path.cwd().joinpath("file_with_names.txt") or "./file_with_names.txt"
     - List[Metabolites]: A list with multiple Metabolite objects.

        Custom metabolite syntax:

        :code:`formatted_identifier, name, compartment, chemical_formula,
        molecular_charge`

        Metabolite from database:

        :code:`metabolite_identifier, compartment`

     - List[str]: A list with multiple str with the mentioned syntax.

     CobraMod will try to download the biochemical information and create
     instead custom metabolites if not found in the database. In case of
     only custom metabolites, it is recommended to use None for the argument
     database

    Args:
        model (Model): Model to be expanded and searched for metabolites.
        obj: A Path; a list with either strings or Metabolite objects,
            or a single string. See syntax above.
        database (str): Name of database. Check
            :obj:`cobramod.available_databases` for a list of names. Defaults
            to None (This is useful for custom metabolites).
        directory (Path): Path to directory where the data is located. If
            nothing is specified, then it defaults to the directory "data" in
            the current working directory.

    Keyword Arguments:
        replacement (dict): Dictionary with either the new identifier and/or
            the identifier of an object inside the model.
        include_metanetx_specific_ec: Determines whether MetaNetX specific
            EC numbers should be taken over. These are generally not found in
            other databases. Furthermore, this could result in non-existing
            Brenda IDs being created. The default value is False.

    Raises:
        WrongSyntax (from str): If the syntax is not followed correctly as
            mentioned above.
        ValueError: If Keyword Arguments are missing.
        FileNotFoundError (from Path): if file does not exists
    """
    # Defaults
    include_metanetx_specific_ec = kwargs.pop(
        "include_metanetx_specific_ec", False
    )
    if not isinstance(directory, Path):
        directory = Path(directory).absolute()

    # Kwargs
    replacement: dict[str, str] = kwargs.pop("replacement", {})
    model_id: str = kwargs.pop("model_id", {})

    if isinstance(obj, str):
        obj = Path(obj).absolute()

    if isinstance(obj, Path):
        metabolites = get_file_metabolites(
            model, obj, replacement, directory, database, model_id
        )

    elif isinstance(obj, list):
        metabolites: list[cobra_core.Metabolite] = []

        for item in obj:
            if isinstance(item, cobra_core.Metabolite):
                metabolites.append(item)

            elif isinstance(item, str):
                metabolites.append(
                    convert_string_metabolite(
                        item, model, replacement, directory, database, model_id
                    )
                )

    else:
        raise cmod_error.WrongDataError(
            "Given object is not the type mentioned in the docstrings. "
            + "Please use a string, a list or a pathlib 'Path' object."
        )

    for metabolite in metabolites:
        crossreferences.add_crossreferences(
            object=metabolite,
            directory=directory,
            include_metanetx_specific_ec=include_metanetx_specific_ec,
        )

    cmod_utils.confirm_metabolite(model, metabolites)


def find_replacements(
    identifier: str,
    obj_type: str,
    replace_dict: dict[str, str],
    model: cobra_core.Model,
) -> Optional[cobra_core.Object]:
    """
    Returns either a COBRApy object if found in the model or a string with the
    new identifier, which can be used to create a new object or retrieve data.
    The replace_dict has two options for keys. One can be the unmodified
    identifier given by a database, or a formated identifier. E.g:
    "WATER_c": "C00001_c" or "ACETALD-DEHYDROG-RXN" : "R00228_c"

    Args:
        identifier (str): Original identifier to replace or rename.
        obj_type (str): Type of object to search for. Only options are
            "reactions" or "metabolites".
        replace_dict (dict): Dictionary with either the new identifier and or
            the identifier of object inside the model.
        model (Model):

    Returns:
        A COBRApy Object if replacement is found in the model. Otherwise
        nothing is returned
    """
    cobra_dictlist: dict[str, cobra_core.DictList] = {
        "reactions": model.reactions,
        "metabolites": model.metabolites,
    }
    # Check if KeyErrors might appear, otherwise continue
    try:
        new_identifier = replace_dict[identifier]
        # Return either COBRApy object or new string to search up in database
        replacement = cobra_dictlist[obj_type].get_by_id(new_identifier)
        if not isinstance(replacement, cobra_core.Object):
            raise TypeError(
                f"The object for {identifier} is not a valid COBRApy object!"
            )

        debug_log.debug(
            f"{obj_type[:-1]} '{identifier}' was found in the model and will "
            " replace {new_identifier}."
        )
        return replacement
    except KeyError:
        return None


def add_reactions(
    model: cobra_core.Model,
    obj: Union[str, Path, list[cobra_core.Object], list[str]],
    directory: Union[Path, str],
    database: Optional[str] = None,
    stop_imbalance: bool = False,
    show_imbalance: bool = True,
    **kwargs,
):
    """Adds given object into the model. The options are:

     - Path: A file with components. E. g. :
        Path.cwd().joinpath("file_with_names.txt")
     - List[Reactions]: A list with regular Reactions
     - List[str]: Either the identifier with its corresponding compartment or a
     string with all components. This applies for the Path option. E.g. :

        :code:`reaction_identifier, compartment`

        For custom reactions

        :code:`reaction_identifier, reaction_name | coefficient metabolite <->
        coefficient metabolite

        Identifiers of metabolites have to end with an underscore and a
        compartment:

            E.g. **`4 OXYGEN-MOLECULE_c`**

     - List[str]: A list with multiple str with the mentioned syntax.

     CobraMod will try to use the same metabolites and reactions inside
     of the model. If not found, then CobraMod will try to download its
     biochemical information or create custom objects.

    Args:
        model (Model): Model to expand and search for reactions.
        obj: A Path; a list with either strings or Reaction objects,
            or a single string. See syntax above.
        database (str): Name of database. Check
            :obj:`cobramod.available_databases` for a list of names. Defaults
            to None (Useful for custom reactions).
        directory (Path): Path to directory where data is located.
        stop_imbalance (bool): If an unbalanced reaction is found,
            stop the process. Defaults to False.
        show_imbalance (bool): If an unbalanced reaction is found, print
            output. Defaults to True.

    Keyword Arguments:
        replacement (dict): Dictionary with either the new identifier and/or
            the identifier of an object inside the model.
        genome (str): Exclusive for KEGG. Abbreviation for the species
            involved. Genes will be obtained for this species.
        consider_sub_elements: Specifies whether additional cross-references
            should also be added to the subelements. For example, you can
            specify whether only the reaction or also its metabolites
            should be expanded. Defaults to True
        include_metanetx_specific_ec: Determines whether MetaNetX specific
            EC numbers should be taken over. These are generally not found in
            other databases. Furthermore, this could result in non-existing
            Brenda IDs being created. The default value is False.

    Raises:
        WrongSyntax (from str): If the syntax is not followed correctly as
            mentioned above.
        ValueError: If Keyword Arguments are missing.
        FileNotFoundError (from Path): If the file does not exists.

    """
    # Defaults
    replacement: dict[str, str] = kwargs.pop("replacement", dict())
    genome: str = kwargs.pop("genome", None)
    model_id: str = kwargs.pop("model_id", None)
    consider_sub_elements = kwargs.pop("consider_sub_elements", True)
    include_metanetx_specific_ec = kwargs.pop(
        "include_metanetx_specific_ec", False
    )

    if isinstance(directory, str):
        directory = Path(directory).absolute()

    if isinstance(obj, str):
        obj = Path(obj).absolute()

    # In case of a Path
    if isinstance(obj, Path):
        reactions = get_file_reactions(
            model=model,
            filename=obj,
            directory=directory,
            database=database,
            replacement=replacement,
            stop_imbalance=stop_imbalance,
            show_imbalance=show_imbalance,
            genome=genome,
            model_id=model_id,
        )

    elif isinstance(obj, list):
        reactions: list[cobra_core.Reaction] = []

        for item in obj:
            if isinstance(item, str):
                reactions.append(
                    string_to_reaction(
                        line=item,
                        model=model,
                        directory=directory,
                        database=database,
                        replacement=replacement,
                        stop_imbalance=stop_imbalance,
                        show_imbalance=show_imbalance,
                        genome=genome,
                        model_id=model_id,
                    )
                )
            if isinstance(item, cobra_core.Reaction):
                reactions.append(item)

    # Raise error if wrong
    else:
        raise cmod_error.WrongDataError(
            "Given object is not the type mentioned in the docstrings. "
            + "Please use a string, a list or a pathlib 'Path' object."
        )

    for reaction in reactions:
        # add cross references only if reactions is non custom
        if not database:
            continue

        crossreferences.add_crossreferences(
            object=reaction,
            directory=directory,
            consider_sub_elements=consider_sub_elements,
            include_metanetx_specific_ec=include_metanetx_specific_ec,
        )

    cmod_utils.add_reactions_to_model(model=model, reactions=reactions)



"""CobraMod Package

CobraMod is a tool for pathway-centric metabolic network curation and extension
based on COBRApy. This package facilitates the retrieval of biochemical data
from multiple databases and help users extend metabolic networks.

CobraMod ensures an easy-friendly API, customization and a record of all the
changes made to metabolic models.

Important functions from CobraMod that can be imported directly:

- get_data: Store and retrieve data from given database.
- create_object: Create a corresponding COBRApy object from given identifier.
- add_metabolites: Add metabolites from different objects
- add_reactions: Add reactions from different objects
- add_pathway: Extend given pathway into given model.
- test_non_zero_flux: Test given reaction to check for a feasible solution.
- add_crossreferences: Check for cross-references.

The exclusive class :func:`cobramod.pathway.Pathway` is an extension of
:class:`cobra.core.group.Group`. CobraMod uses :mod:`Escher` to visualize
the pathway and fluxes of that specific class. Some new methods added:

- visualize: Return a Builder for the representation of the pathway.
- solution: Filters solution and returns fluxes of only members of the class.

By default, CobraMod creates the file "debug.log", which displays the changes,
which occurred when running a script. In order to modify the logging, import
"debug_log" from "cobramod.debug". The default logging level is defined as
INFO. Read the documentation of logging for more information.

For a list of databases, load variable :obj:`cobramod.available_databases`
"""
from cobramod.core.creation import (
    add_metabolites,
    add_reactions,
    create_object,
)
from cobramod.core.crossreferences import add_crossreferences
from cobramod.core.extension import test_non_zero_flux, add_pathway
from cobramod.core.pathway import Pathway, model_convert
from cobramod.retrieval import get_data

__all__ = [
    "get_data",
    "create_object",
    "add_reactions",
    "add_metabolites",
    "add_pathway",
    "test_non_zero_flux",
    "Pathway",
    "model_convert",
    "add_crossreferences",
]

__version__ = "1.2.0"



"""
Debugging configuration for CobraMod

The logs are saved in the directory 'logs' created in the working directory.
Cobramod saves the logs by date, i.e, running the commands in different days
will results in different files. The default level is set to INFO.

The name of logger variable is `debug_log`
"""
import datetime as dt
import logging
from pathlib import Path

import colorlog

debug_log = logging.getLogger("debug_log")
debug_log.setLevel(logging.INFO)

format_str = "%(log_color)s%(message)s"
TIME_STR = "%H:%M:%S"

colors = {
    "DEBUG": "white",
    "INFO": "cyan",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "purple",
}
formatter = colorlog.ColoredFormatter(format_str, TIME_STR, log_colors=colors)

# In direct stream
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
debug_log.addHandler(stream_handler)

# File logs
format_str = "[%(asctime)s] %(levelname)s %(message)s"
formatter = logging.Formatter(format_str, TIME_STR)

log_dir = Path("logs").absolute()
log_dir.mkdir(exist_ok=True)
log_path = log_dir.joinpath(
    f"cobramod_{dt.date.today().strftime('%Y%m%d')}.log"
)

debug_handler = logging.FileHandler(log_path, mode="a+")
debug_handler.setFormatter(formatter)

debug_log.addHandler(debug_handler)


def change_to_debug() -> None:
    """
    Changes the logging level to DEBUG and changes the formatting of it.
    This function is intented to be used only for the tests
    """
    debug_log.setLevel(logging.DEBUG)
    format_str = "[%(asctime)s] %(log_color)s%(levelname)s %(message)s"
    formatter = colorlog.ColoredFormatter(
        format_str, TIME_STR, log_colors=colors
    )
    stream_handler.setFormatter(formatter)



import re
from functools import lru_cache
from pathlib import Path
from typing import Set, Union, List, Any

import pandas as pd
import requests
from cobra import Model, Reaction, Metabolite
from cobra.core import Group
from requests import HTTPError
from tqdm import tqdm

from cobramod.debug import debug_log


def inchikey2pubchem_cid(
    inchikey: Union[str, List[str]], directory: Path
) -> Union[str, List[str]]:
    """
    This function returns the corresponding PubChem compound ID for an
    InChIKey. A local cache is used, which is located in the specified
    directory under the folder XREF.
    Args:
        inchikey: The InChIKey for which the PubChem compound ID is to be
            searched.
        directory: The directory for storing the data. This is where
            the cache is stored in a folder called XRef.

    Returns:
        The PubChem compound ID(s) found as a string if it is one or as a
            list of strings otherwise.
    """
    if isinstance(inchikey, list):
        result: List[str] = []
        for key in inchikey:
            # following can return strings
            result.append(inchikey2pubchem_cid(key, directory))  # type: ignore
        return result

    cache = load_cache_from_disk("pubchem", directory)
    value = cache.loc[cache["ID"] == inchikey]["XRefs"]

    if len(value) > 0:
        return value.iloc[0]

    url = (
        "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/inchikey/"
        + inchikey
        + "/cids/txt"
    )

    response = requests.get(url=url)
    response.raise_for_status()
    value = response.text.rstrip()

    # cache = cache.append({"ID": inchikey, "XRefs": value}, ignore_index=True)
    cache = pd.concat(
        [cache, pd.DataFrame([{"ID": inchikey, "XRefs": value}])],
        ignore_index=True,
    )
    path = directory / "XRef" / str("pubchem" + ".feather")
    path.parent.mkdir(parents=True, exist_ok=True)
    cache.to_feather(path)
    load_cache_from_disk.cache_clear()
    return value


@lru_cache(maxsize=10)
def load_cache_from_disk(sort: str, directory: Path) -> pd.DataFrame:
    """
    The function loads the locally stored cache and returns it as a
    pandas DataFrame. If there is no cache file yet an empty DataFrame
    with columns 'ID' and 'XRefs' is returned.
    Args:
        sort: Der cache Name.
        directory: The directory for storing the data. This is where
            the cache is stored in a folder called XRef.

    """

    try:
        df = pd.read_feather(directory / "XRef" / str(sort + ".feather"))
    except FileNotFoundError:
        df = pd.DataFrame({"ID": [], "XRefs": []})

    return df


def get_crossreferences(  # noqa: C901
    sort: str, querys: Union[str, List[str]], directory: Path
) -> Set[str]:
    """
    Searches for IDs, other IDs from other databases. MetaNetX is
    used for this purpose. Results are stored locally in the specified
    directory and if they are found in it, they are loaded from it.
    Args:
        sort: Type of IDs' possible specifications are "chem" for
            metabolites or "reac" for reactions.
        querys: The IDs of a metabolite or a reaction. Can be either a string
            of the form "database:identifier" or a list of such strings.
            The list should only consist of identifiers for an object,
            as these will be merged.
        directory: The directory for storing the data. This is where
            the cache is stored in a folder called XRef.

    Returns:
        All references retrieved as a set of strings of the
        structure: "database:ID".
    """

    query_list: str
    cache = load_cache_from_disk(sort, directory)
    crossreferences = set()

    if isinstance(querys, list):
        all_querys = querys.copy()
        for query in querys:
            result = cache.loc[cache["ID"] == query]["XRefs"]

            if len(result) > 0:
                all_querys.remove(query)
                crossreferences.update(set(result.iloc[0]))

        if len(all_querys) == 0:
            return crossreferences
        query_list = " ".join(all_querys)
    else:
        all_querys = [querys]
        result = cache.loc[cache["ID"] == querys]["XRefs"]
        if len(result) > 0:
            return set(result.iloc[0])
        query_list = querys

    url = "https://www.metanetx.org/cgi-bin/mnxweb/id-mapper"
    data = {
        "query_list": query_list,
        "query_index": sort,
        "output_format": "json",
    }

    response = requests.post(url=url, data=data)
    try:
        response.raise_for_status()
    except HTTPError as error:
        if error.response.status_code == 413:
            size = len(querys)
            half = round(size / 2)
            chunk1 = querys[1:half]
            chunk2 = querys[half + 1 : size]

            result1 = get_crossreferences(sort, chunk1, directory)
            result2 = get_crossreferences(sort, chunk2, directory)

            return set.union(result1, result2)
        else:
            return set()
    response_json = response.json()

    for query in all_querys:
        try:
            answer = response_json[query]
        except KeyError:
            continue

        xrefs: List[str] = []
        try:
            xrefs = answer["xrefs"]
        except KeyError:
            pass
        try:
            xrefs.append("inchi:" + answer["InChI"])
        except KeyError:
            pass
        try:
            inchikey = answer["InChIkey"]
            xrefs.append("inchikey:" + inchikey)
        except KeyError:
            pass

        replace = {"chem": "metanetx.chemical", "reac": "metanetx.reaction"}

        try:
            mnx_id = answer["mnx_id"]
            xrefs.append(replace[sort] + ":" + mnx_id)
        except KeyError:
            pass

        cache = pd.concat(
            [cache, pd.DataFrame([{"ID": query, "XRefs": xrefs}])],
            ignore_index=True,
        )
        # cache = cache.append({"ID": query, "XRefs": xrefs}, ignore_index=True)
        path = directory / "XRef" / str(sort + ".feather")
        path.parent.mkdir(parents=True, exist_ok=True)
        cache.to_feather(path)
        load_cache_from_disk.cache_clear()
        crossreferences.update(xrefs)

    return crossreferences


def metanetx2ec(
    id: Union[str, List[str]],
    directory: Path,
    include_metanetx_specific_ec: bool = False,
) -> Union[str, List[str]]:
    """
    Returns the corresponding EC number for a specific MetaNetX ID.
    Args:
        id: The MetaNetX ID.
        directory: The directory for storing the data. This is where
            the cache is stored in a folder called XRef.
        include_metanetx_specific_ec: Determines whether MetaNetX specific
            EC numbers should be taken over. These are generally not found in
            other databases. The default value is False.
    Raises
        KeyError: If no EC number can be assigned to this ID.
    Returns:
        All found EC numbers.
    """
    if isinstance(id, List):
        result = set()
        for single_id in id:
            try:
                ec_numbers = metanetx2ec(
                    single_id, directory, include_metanetx_specific_ec
                )
            except KeyError:
                continue

            if isinstance(ec_numbers, str):
                ec_numbers = [ec_numbers]
            result.update(ec_numbers)

        size = len(result)
        if size == 0:
            raise KeyError
        elif size == 1:
            return result.pop()
        else:
            return list(result)

    data = get_reac_prop_with_ec(directory)
    found_ec_numbers = data.loc[data["ID"] == id, "classifs"]
    pattern = re.compile(
        r"^\d+\.-\.-\.-|\d+\.\d+\.-\.-|\d+\."
        r"\d+\.\d+\.-|\d+\.\d+\.\d+\.(n)?\d+$"
    )

    if len(found_ec_numbers) == 0:
        raise KeyError

    found_ec_numbers = found_ec_numbers.iloc[0]

    if ";" in found_ec_numbers:
        found_ec_numbers = found_ec_numbers.split(";")

        for found_ec_number in found_ec_numbers:
            # id can only be a string at this point
            if not include_metanetx_specific_ec and not pattern.match(
                found_ec_number
            ):
                found_ec_numbers.remove(found_ec_number)

        if len(found_ec_numbers) == 0:
            raise KeyError

    else:
        if not include_metanetx_specific_ec and not pattern.match(
            found_ec_numbers
        ):
            raise KeyError

    return found_ec_numbers


@lru_cache(maxsize=1)
def get_reac_prop_with_ec(directory: Path) -> pd.DataFrame:
    """
    This function loads the file reac_prop from MetaNetX and stores it in
    memory using the lru_cache. The returned DataFrame contains only the
    rows that have an EC number.

    Args:
        directory: The directory for storing the data. This is where
            the cache is stored in a folder called XRef.
    """

    path = directory / "XRef" / str("reac_prop" + ".feather")

    try:
        return pd.read_feather(path)
    except FileNotFoundError:
        pass

    url = "https://www.metanetx.org/ftp/latest/reac_prop.tsv"

    df = pd.read_csv(
        url,
        sep="\t",
        comment="#",
        names=[
            "ID",
            "mnx_equation",
            "reference",
            "classifs",
            "is_balanced",
            "is_transport",
        ],
    )

    df = df[~df["classifs"].isna()].reset_index(drop=True)

    path.parent.mkdir(parents=True, exist_ok=True)
    df.to_feather(path)

    return df


def add2dict_unique(
    key, value: Union[Any, List[Any]], dictionary: dict
) -> dict:
    """
    Adds key-value pairs to a dictionary. It is expected that the values
    are either single objects or in a list. If the key does not yet exist
    in the dictionary, the value is added directly with it. If the key
    already exists, the values are joined together like this. That a list
    is assigned to the key, which contains unique values. If only one value
    remains, it will be added directly and not in a list.
    Args:
        key: The key to be used.
        value: The Values to be added. This can be a single object or
            objects in a list.
        dictionary: The dictionary to which the objects should be added.

    Returns:
        The new dictionary.
    """

    if key not in dictionary.keys():
        dictionary[key] = value
    else:
        ref = dictionary[key]
        if isinstance(ref, str):
            ref = [ref]
        if isinstance(value, str):
            value = [value]

        unique_pub = list(set(ref + value))
        if len(unique_pub) == 1:
            dictionary[key] = unique_pub[0]
        else:
            dictionary[key] = unique_pub

    return dictionary


def add_crossreferences(  # noqa: C901
    object: Union[Model, Group, Reaction, Metabolite],
    directory: Union[Path, str],
    consider_sub_elements: bool = True,
    include_metanetx_specific_ec: bool = False,
) -> None:
    """Extends the passed object by cross-references. Here, only the
    cross-references of reactions or metabolites are expanded. There
    must be at least one MetaNetX resolvable identifier in the annotation
    of the object in order to get as many cross-references as possible.
    Overall, an attempt is made to add all cross-references required by Memote.
    The only exception are reactom reactions. These are not added due to the
    current lack of query capabilities on the part of reactom. (Depending on
    the number of objects this function can take some time).

    Args:
        object: The CobraPy object to be extended.
        directory: The directory for storing the data. This is where
            the cache is stored in a folder called XRef.
        consider_sub_elements: Specifies whether additional cross-references
            should also be added to the subelements. For example, you can
            specify whether only the reaction or also its metabolites
            should be expanded.
        include_metanetx_specific_ec: Determines whether MetaNetX specific
            EC numbers should be taken over. These are generally not found in
            other databases. Furthermore, this could result in non-existing
            Brenda IDs being created. The default value is False.

    """
    if isinstance(directory, str):
        directory = Path(directory).absolute()

    if isinstance(object, Model):
        if not consider_sub_elements:
            debug_log.warn(
                "A model was passed but no sub-elements are to be annotated. "
                "Nothing changes."
            )
            return
        size = len(object.reactions) + len(object.metabolites)
        debug_log.debug(
            f"A model was passed. Trying to find further references for "
            f"{size} elements."
        )
        with tqdm(total=size) as pbar:
            for reaction in object.reactions:
                add_crossreferences(
                    reaction, directory=directory, consider_sub_elements=False
                )
                pbar.update(1)

            for metabolite in object.metabolites:
                add_crossreferences(metabolite, directory)
                pbar.update(1)

    elif isinstance(object, Group):
        if not consider_sub_elements:
            debug_log.warn(
                "A group was passed but no sub-elements are to be annotated. "
                "Nothing changes."
            )
            return

        debug_log.debug(
            f"A group was passed. Trying to find further references for "
            f"{len(object.members)} elements."
        )
        for member in tqdm(object.members):
            add_crossreferences(
                member, directory=directory, consider_sub_elements=False
            )

    else:
        sort = None

        if isinstance(object, Reaction):
            if consider_sub_elements:
                for metabolite in object.metabolites:
                    add_crossreferences(metabolite, directory)

            sort = "reac"

        elif isinstance(object, Metabolite):
            sort = "chem"

        if sort is None:
            debug_log.error(
                "An object was passed that is not of type Model, Group, "
                "Reaction or Metabolite! Check the 'object' value."
            )
            raise ValueError

        # metanetx

        if object.id is not None:
            id = object.id
            ids = [object.id]
        else:
            id = "Undefined"
            ids = []
        xrefs = object.annotation
        size = len(xrefs)
        total_found = 0

        for key, value in object.annotation.items():
            if isinstance(value, list):
                for id in value:
                    ids.append(key + ":" + id)
            else:
                ids.append(key + ":" + value)

        potential_xrefs = get_crossreferences(sort, ids, directory)

        for potential_xref in potential_xrefs:
            try:
                prefix, new_id = potential_xref.split(":")
            except ValueError:
                continue
            prefix = prefix.lower()

            # MetaNetX does not provide correct CHEBI IDs. This is because
            # they lack the correct prefix. "CHEBI:CHEBI:0000" is correct
            # but "CHEBI:0000" is delivered. Therefore the following if
            # statement is used. This will also work if this condition is
            # fixed. (But then it is redundant)
            if prefix == "chebi" and "CHEBI:" not in new_id:
                new_id = "CHEBI:" + new_id

            # obsolete ids are ignored
            elif prefix == "deprecated":
                continue

            total_found += 1
            xrefs = add2dict_unique(prefix, new_id, xrefs)

        # pubchem.compound

        try:
            inchikey = xrefs["inchikey"]
            pubchem_compound = inchikey2pubchem_cid(inchikey, directory)
        except (KeyError, HTTPError):
            pass
        else:
            xrefs = add2dict_unique("pubchem.compound", pubchem_compound, xrefs)

            if isinstance(pubchem_compound, str):
                pubchem_compound = [pubchem_compound]

            total_found += len(pubchem_compound)

        # ec_number
        try:
            metanet_xid = xrefs["metanetx.reaction"]
            ec_number = metanetx2ec(
                metanet_xid,
                directory,
                include_metanetx_specific_ec=include_metanetx_specific_ec,
            )
        except KeyError:
            pass
        else:
            xrefs = add2dict_unique("ec-code", ec_number, xrefs)

            if isinstance(ec_number, str):
                ec_number = [ec_number]

            total_found += len(ec_number)

        # brenda
        try:
            ec_number = xrefs["ec-code"]
        except KeyError:
            pass
        else:
            xrefs = add2dict_unique("brenda", ec_number, xrefs)

            if isinstance(ec_number, str):
                ec_number = [ec_number]

            total_found += len(ec_number)

        total_added = len(xrefs) - size

        debug_log.debug(
            f'For the object with the ID {id} a total of "{total_found}" '
            f"references were found. Of these, {total_added} "
            f"were missing and have been added."
        )

        object.annotation = xrefs



"""Module for graph algorithm

This module creates the needed functions to find out the reaction order from
a pathway. The vertex represents the reactions and the edges symbolize the
order of the reactions. I.e. the relationship between reactions. The main
function of this module:

build_graph: From given dictionary with Parent-reaction:children-reaction,
return the corresponding non-cyclic directed graph.

"""
from __future__ import annotations

from collections import Counter
from contextlib import suppress
from itertools import chain
from typing import Any, Dict, Union

from cobramod.error import GraphKeyError


def find_missing(graph: dict[str, Union[str, tuple[str]]]):
    """
    Checks whether the given graph is missing a key.

    Args:
        graph (dict): Dictionary representing the relationships between nodes.
            A node can have several edges, which should be represented in the
            form of values.
    Raises:
        KeyError: If keys are missing
    """
    values = set()
    for value in graph.values():
        if isinstance(value, tuple):
            for item in value:
                values.add(item)
            continue
        values.add(value)
    keys = {key for key in graph.keys()}
    intersection = set.difference(values, keys)
    intersection = set(filter(None, intersection))
    if len(intersection) > 0:
        raise GraphKeyError(
            f"The graph is missing following keys: {intersection}"
        )


def find_cycle(
    graph: dict[str, Union[str, tuple[str]]], key: str, visited: list[str]
) -> list[str]:
    visited.append(key)

    value = graph.get(key, None)

    if isinstance(value, str):
        if value in visited:
            return visited

        return find_cycle(graph, value, visited)

    elif isinstance(value, tuple):
        from_tuple: list[list[str]] = list()

        for element in value:
            # In case of a set, all values must be tested as well
            old_visited = visited.copy()

            branch = find_cycle(graph, element, old_visited)

            if branch:
                from_tuple.append(branch)

        if from_tuple:
            return from_tuple.pop()

    return []


def return_cycles(graph: dict[str, Union[str, tuple[str]]]):
    """
    Returns a nested list of cyclic paths. These paths might repeat. If the
    graph does not contain a cycle, the list is empty.

    Args:
        graph (dict): Dictionary representing the relationships between nodes.
            A node can have several edges, which should be represented in the
            form of values.

    Returns:
        List: Nested list with cyclic paths or an empty list if the graph
            does not contain a cycle.
    """
    cycles: list[list[str]] = list()

    for node in graph.keys():
        cycle = find_cycle(graph=graph, key=node, visited=[])

        if cycle:
            cycles.append(cycle)

    return cycles


def cut_cycle(graph: dict[str, Union[str, None, tuple[str]]], key: str):
    """
    Changes value of the key to None in a given dictionary. It will raise an
    error if the value is a tuple.
    """
    if not isinstance(key, str):
        raise AttributeError(
            f'Given key "{key}" cannot be cut. Probably because it is a tuple'
        )
    graph[key] = None


def cut_parents(graph: dict):
    """
    Checks if multiple parents shared a common child. If so, the graph will
    replace the values of these parents to a None and leave one of the parents
    normal.


    Args:
        graph (dict): Dictionary representing the relationships between nodes.
            A node can have several edges, which should be represented in the
            form of values.
    """
    counter = Counter(graph.values())
    # Get the keys that are not one
    values = [key for key, value in counter.items() if value != 1]
    # Get candidates
    candidates: Dict[str, Any] = dict()
    for key, value in graph.items():
        if value in values:
            try:
                candidates[value].append(key)
            except KeyError:
                candidates[value] = [key]
    # Cut all candidates but the first item
    for values in candidates.values():
        for index, key in enumerate(values):
            if index == 0:
                continue
            graph[key] = None


def back(
    graph: dict[str, Union[str, tuple[str]]],
    value: str,
    path: list[str],
    stop_list: list = [],
) -> list[str]:
    """
    Returns a list with a linear path. The function creates a sequence from
    the graph dictionary until the given value is found in the sequence or in
    the given stop_list. The function does not work with graphs that contain
    some kind of cycle and will raise a recursion error.

    Args:
        graph (dict): Dictionary representing the relationships between nodes.
            A node can have several edges, which should be represented in the
            form of values.
        value (str): The value to be searched..
        path (list): The already-visited path.
        stop_list (list): Elements that trigger the function to stop, if found.

    Returns:
        List: A list with the path that ends with the specified value or with
            an element of the stop_list.

    Raises:
        RecursionError: If an element is visited more than once due to a cycle.
    """
    # Check for each key if the value matches the argument value. Extend path
    # and call recursive
    for key, val in graph.items():
        if isinstance(val, tuple):
            # In case of tuples, test each single case
            for single in val:
                if single in stop_list:
                    path.insert(0, key)
                    return path
                if single == value:
                    path.insert(0, key)
                    return back(
                        graph=graph, value=key, path=path, stop_list=stop_list
                    )
        if val in stop_list:
            path.insert(0, key)
            return path
        # If val from dictionary is the same as value, then run recursive.
        if val == value:
            path.insert(0, key)
            return back(graph=graph, value=key, path=path, stop_list=stop_list)
    return path


def verify_paths(paths: list, graph: dict) -> set[str]:
    """
    Returns the missing keys that are not present in the given paths list.

    Args:
        paths (list): Paths of the given graph.
        graph (dict): Dictionary representing the relationships between nodes.
            A node can have several edges, which should be represented in the
            form of values.

    Returns:
        set: Either the missing nodes or an empty set.
    """
    reactions = set(chain.from_iterable(paths))
    return set(graph.keys()).difference(reactions)


def get_paths(graph: dict, stop_list: list) -> list[list[str]]:
    """
    Returns a list with the longest paths in a graph. This only works with
    lineal directed paths.

    Args:
        graph (dict): Dictionary representing the relationships between nodes.
            A node can have several edges, which should be represented in the
            form of values.

    Returns:
        List: Paths from given graph

    Raises:
        RecursionError: if the graph is not lineal
    """
    end_nodes = [key for key, value in graph.items() if value is None]

    # Function back will raise RecursionError if not lineal
    path_generator = (
        back(graph=graph, value=node, path=[node], stop_list=stop_list)
        for node in end_nodes
    )
    paths: list[list[str]] = list()

    for _ in end_nodes:
        with suppress(StopIteration):
            single = next(path_generator)
            paths.append(single)

    # FIXME: this is a temporal solution. When multiple tuples are located in
    # the graph, back would not be able to find all member
    difference = verify_paths(paths=paths, graph=graph)
    if difference:
        # Must be independent of each other, otherwise, the order might change
        # the behavior of the visualization
        for item in difference:
            paths.append([item])

    return paths


def get_mapping(graph: dict, stop_list: list, new: list) -> list:
    """
    Gets the mapping for given graph. The mapping defines the longest paths
    for the graph without including the previous longest path. The function
    modifies given graph.

    Args:
        graph (dict): graph (dict): Dictionary representing the relationships
            between nodes. A node can have several edges, which should be
            represented in the form of values. This will be modified.
        stop_list (list): Elements that cause the function to stop when found.
        new (list): List with new mapping. Must be empty when called for
            the first time.

    Returns:
        List: A list with the new mapping. Longest path for each recursion and
            rest.
    """
    paths = get_paths(graph=graph, stop_list=stop_list)
    # Get longest and reduce graph
    try:
        longest = max(paths, key=len)
    except ValueError:
        return []
    new.append(longest)
    for item in longest:
        graph.pop(item)
    get_mapping(graph=graph, stop_list=longest, new=new)
    return new


def get_graph_dict(graph: dict[str, Union[str, tuple[str]]]) -> list[list[str]]:
    """
    Returns the mapping for the given graph. The mapping is defined as a list
    with a "core" path and its branches. Cyclic graphs will be cut to create
    a lineal direct graph

    .. note::
        If the first element of a branch is not in the "core", then it is
        in one of the branches.

    Returns:
        List: Mapping of the graph.

    Raises:
        GraphKeyError: If graph is missing a value.
    """
    # Check that all values are represented
    find_missing(graph=graph)
    # its value cannot be None. If single element then go directly to mapping
    with suppress(KeyError):
        # Cut parents if needed
        cut_parents(graph=graph)
        # Fix cycles if found
        cycles = return_cycles(graph=graph)
        # Check until no cycles are found.
        while cycles:
            # This is modify graph
            # TODO: check whether tuples are affected
            cut_cycle(graph=graph, key=cycles[0][0])
            cycles = return_cycles(graph=graph)
    # This would modify the graph. Use copy
    mapping = get_mapping(graph=graph.copy(), stop_list=[], new=[])
    mapping.sort(key=len, reverse=True)
    return mapping


def filter_graph(graph: dict, avoid_list: list) -> dict:
    """
    Returns a new graph, where items are removed if found in given avoid list.
    """
    # TODO: docstrings
    new_graph = graph.copy()
    # Remove necessary keys in copy
    for reaction in avoid_list:
        with suppress(KeyError):
            del new_graph[reaction]
        for key, value in new_graph.items():
            # Avoid NoneType

            if not value:
                continue

            elif reaction in value and isinstance(value, tuple):
                new_graph[key] = tuple(
                    item for item in value if item != reaction
                )
                # Transform to str if only one element
                if len(new_graph[key]) == 1:
                    new_graph[key] = new_graph[key][0]

            elif reaction == value:
                new_graph[key] = None
    return new_graph


def build_lineal_graph(sequence: list[str]) -> dict[str, Union[str, None]]:
    """
    Creates a returns a simple lineal directed graph from given sequence
    """
    graph: dict[str, Union[None, str]] = dict()
    for index, reaction in enumerate(sequence):
        try:
            parent = reaction
            child = sequence[index + 1]
            graph[parent] = child

        except IndexError:
            # It must be the first one
            graph[reaction] = None

    return graph



"""JSON creator

This module handles the convertion of strings into a proper object that can be
used to create a JSON string, which can be later used in Escher.

The main class of this module is
:class:`cobramod.visualization.converter.JsonDictionary`. This class is able to
parse and store data as JSON objects. To check the attributes for each
JSON object, please read the documentation of
:mod:`cobramod.visualization.items`.

Important methods:


- json_dump: The data can be parsed into a JSON. (str)
- add_metabolite: Add metabolite-node into the JsonDictionary.
- add_marker: Add a marker-node into the JsonDictionary.
- create_reaction: Returns a :class:`cobramod.visualization.items.Reaction`.
- add_reaction: Parses a reaction string and add the information into the
JsonDictionary.
- add_blank: Adds a empty reaction. This is useful for the extra space in the
visualizations.
- visualize: Saves Escher visualization as a HTML and return the Escher
Builder.
"""
import fileinput
import math
from collections import UserDict, namedtuple
from contextlib import suppress
from itertools import cycle
from json import dumps
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import webcolors

try:
    import escher
except ImportError:
    # NOTE: there might be a better way
    pass

from cobramod.visualization.debug import debug_log
from cobramod.visualization.items import Node, Reaction
from cobramod.visualization.mapping import get_mapping, transpose

Position = namedtuple("Position", ["row", "column"])


def _convert_string(string: str) -> dict:
    """
    Converts a :class:`cobra.Reaction` reaction-string into a dictionary, and
    returns a dictionary with the corresponding participants and their
    coefficients.

    Examples:
    'C00001_c + 2 C00002_c --> C00009_c + C00080_c + G11113_c'
    'C00002_c + C00033_c <=> C00227_c + G11113_c'
    'C00002_c + C00033_c <-- C00227_c + G11113_c'
    """
    # TODO: return reversibility
    middle = max(string.find(">"), string.find("<"))
    # find exact middle
    if " " == string[middle - 1]:
        middle = middle + 1
    else:
        middle = middle - 1
    left, right = string[: middle - 1], string[middle + 2 :]
    metabolites = dict()
    for side in (left, right):
        # If split do not work, then add a "1"
        sequence = [
            item.split() if len(item.split()) == 2 else ["1", *item.split()]
            for item in side.split("+")
        ]
        # FACTOR defines product or reactant
        FACTOR = 1
        if side is left:
            FACTOR = -1
        # Sometimes, there are no metabolites (sink reactions), the single "1"
        # must be ignored
        with suppress(ValueError):
            metabolites.update(
                {key: float(value) * FACTOR for value, key in sequence}
            )
    return metabolites


def _color2np_rgb(color: Union[str, List[int], None]) -> np.ndarray:
    """
    This function translates the rgb or string representation into a numpy
    array. The string representation must follow the css standard.

    Args:
        color: The color to be translated.

    Returns:
        Numpy array with the rgb representation of the color or with the
        values [220,220,220] as standard return for incorrect inputs.
    """
    try:
        assert color is not None
        if not all(isinstance(elements, int) for elements in color):
            color = webcolors.name_to_rgb(color)
    except (KeyError, TypeError, ValueError, AssertionError):
        debug_log.warning(
            f'Unknown color or wrong format: "{color}". Using default color.'
        )
        color = [220, 220, 220]

    # turn int arrays into numpy arrays
    return np.array(color, dtype=np.float32)


def _divide_values(
    flux: List[float], min_max: Optional[List[float]] = None
) -> Tuple[List[float], List[float], bool, bool]:
    """
    This function divides a list into two, one consisting of the positive
    values and one of the negative values.

    Args:
        flux (List[int]): Array that is to be divided into positive and
            negative values.
        min_max ([int,int]): List consisting of two values. These values
            determine the maximum value and minimum value that are taken into
            account in the distribution. All values outside this interval
            are ignored.
    Returns:
        This function returns two lists and two bools. The first return is the
            list consisting of positive values. The second return is the list
            with negative values. The third return is a bool that indicates
            whether min_max consists of two positive values. The fourth
            return describes whether min_max consists of two negative values.
    """

    # if required, take min and max settings into account and add them if
    # not present

    both_positive = False
    both_negative = False

    for _ in [0]:
        if min_max is None:
            continue

        if min_max[0] > min_max[1]:
            debug_log.warn(
                "Set minimum is greater than maximum. Ignoring min_max"
            )
            continue

        both_positive = min_max[0] > 0 and min_max[1] > 0
        both_negative = min_max[0] < 0 and min_max[1] < 0

        size = len(flux)
        msg = (
            f"Due to set min_max values were ignored. Original range "
            f"was [{min(flux)},{max(flux)}] "
            f"set is [{min_max[0]},{min_max[1]}]."
        )

        flux = [value for value in flux if min_max[0] < value < min_max[1]]
        if size > len(flux):
            debug_log.info(msg)

        if min_max[0] != 0 and min_max[0] not in flux:
            flux.append(min_max[0])

        if min_max[1] != 0 and min_max[1] not in flux:
            flux.append(min_max[1])

    # divide positive and negative values
    positive = list()
    negative = list()

    for value in flux:
        if value > 0:
            positive.append(value)
        elif value < 0:
            negative.append(value)

    return positive, negative, both_positive, both_negative


class JsonDictionary(UserDict):
    """
    Create a JsonDictionary object which can be used to parse information into
    JSON files to be later read by Escher. When creating the object, some
    keyword arguments may be passed.

    Keyword Arguments:
        head (dict, optional): General information of the JSON. Keys
            included: map_name, map_id, map_description, homepage, schema.
        reactions (dict, optional): Dictionary with multiple
            :class:`cobramod.visualization.items.Reaction` where the key is the
            number of the reaction and the value the object. Defaults to empty
            dictionary.
        nodes (PairDictionary): Dictionary with multiple
            :class:`cobramod.visualization.items.Node` where the key is the
            number of the Node and the value the corresponding object. Defaults
            to empty dictionary.
        text_labels  (dict, optional): Dictionary with the custom text inside
            the canvas.
        canvas (dict, optional): x and y position, width and height of the
            white area in Escher.

    Attributes:
        CANVAS_WIDTH (float): Width for the canvas. Defaults to 1500.
        CANVAS_HEIGHT (float): Height for the canvas. Defaults to 1500.
        R_WIDTH (float): Width of a reaction. Defaults to 350.
        R_HEIGHT (float): Height of a reaction. Defaults to 270.
        reaction_data (Dict[str, float]): Dictionary with the solution to be
            visualized. Default to empty dictionary.
    """

    def __init__(self):
        """
        Initiates the creation of the JsonDictionary. It uses the same
        arguments and keyword arguments as a regular dictionary. However, some
        important keys can be imported, shown in the docstring for the class.
        """
        # Init dictionary
        super().__init__()
        # Initiate basic information
        self.data["head"] = {
            "map_name": "",
            "map_id": "",
            "map_description": "",
            "homepage": "",
            "schema": "https://escher.github.io/escher/jsonschema/1-0-0#",
        }
        self.data["reactions"] = {}
        # TODO: add PairDictionary. Check if even needed
        self.data["nodes"] = {}
        self.data["text_labels"] = {}
        # Canvas variables
        self.CANVAS_WIDTH: float = 0
        self.CANVAS_HEIGHT: float = 0
        # TODO: fix variables
        self.X = 0
        self.Y = 0
        # Reaction size
        self.R_WIDTH: float = 650
        self.R_HEIGHT: float = 450  # 210
        # Data stored about reactions and participants.
        self._overview = dict()
        # Default solution
        self.flux_solution: Dict[str, float] = None
        # Dictionary with relationship of reactions
        self.graph: dict = dict()
        self.reaction_strings = dict()
        self.reaction_scale = dict()

    def get_canvas(self) -> dict:
        return {
            "x": self.X,
            "y": self.Y,
            "width": self.CANVAS_WIDTH,
            "height": self.CANVAS_HEIGHT,
        }

    def json_dump(self, indent: int = None) -> str:
        """
        Returns a string that is the JSON representation of this class.

        Args:
            indent (int): Defines the indentation for the JSON.
                Defaults to None.
        """
        # Transform into regular dictionaries
        nodes = {
            key: dict(**value) for key, value in self.data["nodes"].items()
        }.copy()
        # In case of reactions, the objects inside them 'Segment' must be
        # transformed as well
        try:
            reactions = {
                key: dict(**value)
                for key, value in self.data["reactions"].items()
            }.copy()
            # Each reaction must have its Segment changed to a native dict
            for reaction in reactions:
                # if empty, must be changed to regular dictionary
                if not reactions[reaction]["segments"]:
                    reactions[reaction]["segments"] = dict()
                else:
                    # Change each Segment
                    temporal_dict = dict()
                    for key, value in reactions[reaction]["segments"].items():
                        temporal_dict[key] = dict(**value)
                    reactions[reaction]["segments"] = temporal_dict

        except KeyError:
            reactions = {}
        return dumps(
            obj=[
                self.data["head"],
                {
                    "reactions": reactions,
                    "nodes": nodes,
                    "text_labels": self.data["text_labels"],
                    "canvas": self.get_canvas(),
                },
            ],
            indent=indent,
        )

    def _get_set(self, item: str) -> set:
        """
        Return a set for the keys of either the reactions, the nodes or
        segments for all reactions. Options for item: "nodes", "segments",
        "reactions".
        """
        if item == "segments":
            numbers = set()
            # Get the numbers from the segments
            with suppress(KeyError):
                reactions = self.data["reactions"]
                for key in reactions.keys():
                    segments = reactions[key]["segments"]
                    numbers.update([int(index) for index in segments])
        elif item in ("reactions", "nodes"):
            numbers = {int(key) for key in self.data[item]}
        else:
            raise ValueError("Argument 'item' not correct. Refer to docstring.")
        return numbers

    def _get_last_number(self, item: str) -> int:
        """
        Returns the largest number of the keys from either reactions, nodes, or
        segments from each reaction. Options for item: "nodes", "segments",
        "reactions".
        """
        # Return 0 for first item, otherwise the longest number + 1
        numbers = self._get_set(item=item)
        if not numbers:
            return 0
        return max(numbers) + 1

    def add_metabolite(
        self,
        x: float,
        y: float,
        label_x: float,
        label_y: float,
        bigg_id: str,
        name: str,
        node_is_primary: bool,
    ):
        """
        Add a metabolite-type node into the JsonDictionary. The key will be
        always the last node number.

        Args:
            x (float): Position in x-axis for the node.
            y (float): Position in y-axis fot the node.
            label_x (float): Position of the label in the x-axis.
            label_y (float): Position of the label in the y-axis.
            bigg_id (str): Identifier of the metabolite. It does not have to be
                from BIGG.
            name (str): Name of the label for the metabolite.
            node_is_primary (bool): True if node should represent a primary
                metabolite, i.e. Node is visually larger.

        """
        number = str(self._get_last_number(item="nodes"))
        self.data["nodes"][number] = Node(
            node_type="metabolite",
            x=x,
            y=y,
            bigg_id=bigg_id,
            label_x=label_x,
            label_y=label_y,
            name=name,
            node_is_primary=node_is_primary,
        )
        debug_log.info(
            f'New metabolite-node for "{bigg_id}" with id "{number}" added'
            f" to the JsonDictionary."
        )

    def add_marker(self, x: float, y: float, node_type: str):
        """
        Add a marker-type node into the JsonDictionary. Node can be a midmarker
        or a multimarker. These markers are located in the middle of the
        reaction. A midmarker is located between two multimarkers.

        Args:
            x (float): Position in x-axis for the node.
            y (float): Position in y-axis for the node.
            node_type (str): Type of marker. Options: 'midmarker' or
                'multimarker'.
        """
        number = str(self._get_last_number(item="nodes"))
        self.data["nodes"][number] = Node(node_type=node_type, x=x, y=y)
        debug_log.info(
            f'New {node_type}-node with id "{number}" added to '
            f"the JsonDictionary"
        )

    def _add_reaction_markers(
        self,
        identifier: str,
        top_edge: float,
        left_edge: float,
        vertical: bool,
    ):
        """
        Add the corresponding midmarker and multimarkers into the
        JsonDictionary. Nodes will be added to the corresponding reaction data.
        Number for the nodes will not repeat themselves. If vertical is passed,
        then the reaction markers will be created in order to represent a
        vertical pathway.
        """
        # For markers: 20 px between each one. Sequence should follow:
        # multimarker-midmarker-multimarker
        # if top_edge is None and left_edge is None:
        #     top_edge, left_edge = self._get_edges()
        for node_type, extra_x, node_position in (
            ("multimarker", -20, "_first"),
            ("midmarker", 0, "_middle"),
            ("multimarker", 20, "_last"),
        ):
            last = self._get_last_number(item="nodes")
            self._overview[identifier]["nodes"][node_position] = last
            # Defining position of labels depending of position axis.
            if vertical:
                self.add_marker(
                    x=left_edge + (self.R_WIDTH / 2),
                    y=top_edge + (self.R_HEIGHT / 2) + extra_x + 0,
                    node_type=node_type,
                )
            else:
                self.add_marker(
                    x=left_edge + (self.R_WIDTH / 2) + extra_x + 30,
                    y=top_edge + (self.R_HEIGHT / 2),
                    node_type=node_type,
                )

    def _get_matrix_reactions(self, vertical: bool, position: int) -> list:
        """
        Return a List with the names of the reactions, which are located in
        given position. The position can be from a column or a row. Vertical
        defines the orientation.
        """
        if not vertical:
            data = [
                reaction
                for reaction in self._overview.keys()
                if self._overview[reaction]["position"].column == position
            ]
        else:
            data = [
                reaction
                for reaction in self._overview.keys()
                if self._overview[reaction]["position"].row == position
            ]
        return data

    def _get_products(self, reactions: list) -> Dict[str, list]:
        """
        Returns a dictionary where keys are the reactions of given list and the
        values the products of the reactions.

        Args:
            reactions (list): List with reaction identifiers.

        Returns:
            dict: Dictionary with products for given reactions.
        """
        previous = dict()
        # Find in class for the Reaction object to find their metabolites
        for reaction in self._overview.keys():
            if reaction in reactions:
                index = self._overview[reaction]["index"]
                try:
                    reaction_obj: Reaction = self.data["reactions"][index]
                    # Obtaining product side of the reactions
                    # key metabolites returns a list
                    previous[reaction] = [
                        metabolite["bigg_id"]
                        for metabolite in reaction_obj["metabolites"]
                        if metabolite["coefficient"] > 0
                    ]
                except Exception:
                    pass
        return previous

    def _find_shared(self, metabolite: str, products: Dict[str, list]) -> tuple:
        """
        Returns the node number of the metabolite and the reaction involved if
        found in given dictionary with products.

        Args:
            metabolite (str): Identifier of the metabolite.
            products (dict): Dictionary with reactions and their corresponding
                products as values.

        Returns:
            tuple: the node_number for the metabolite and the name of the
                involved reaction.
        """
        # If nothing is found, empty strings
        node_number = str()
        old_reaction = str()
        for reaction, metabolites_list in products.items():
            if metabolite in metabolites_list:
                node_number = self._overview[reaction]["nodes"][metabolite]
                old_reaction = reaction
                # Only one is necessary
                break
        return node_number, old_reaction

    def map_metabolites(
        self,
        metabolite_dict: dict,
        reaction: Reaction,
        top_edge: float,
        left_edge: float,
        vertical: bool,
    ):
        """
        Creates the metabolites from given dictionary and complements the
        :class:`cobramod.visualization.items.Reaction`. Moreover, it creates
        the corresponding metabolites-nodes for the JsonDictionary class.

        Args:
            metabolite_dict (dict): Dictionary with metabolites and their
                coefficients.
            reaction (Reaction): Reaction that will include the metabolite.
            top_edge (float): Position for the top edge of the reaction-box.
            left_edge (float): Position for the left edge of the reaction-box.
            vertical (bool): Defines if metabolite should be map vertically.
        """
        # Obtaining position of reaction
        identifier = reaction["bigg_id"]
        position: Position = self._overview[identifier]["position"]
        # Minimum number of identifiers. TODO: verify behavior with 0
        side_dict = {"left": 1, "right": 1}
        # Obtain products from reactions in the prior column.
        # The column must be actual - 1, or 0. This is to check for shared
        # metabolites.
        position_value = position.column
        if vertical:
            position_value = position.row
        # Obtains reactions that shared either the previous column or row
        shared_reactions = self._get_matrix_reactions(
            vertical=vertical, position=max(position_value - 1, 0)
        )
        # Remove actual identifier
        filtered = [
            reaction for reaction in shared_reactions if reaction != identifier
        ]
        shared_products = self._get_products(reactions=filtered)
        # Add metabolites from dictionary.
        # TODO: get rid off if-statements
        for metabolite, coefficient in metabolite_dict.items():
            # By default, left side
            item = "reactants"
            SIDE = 0
            counter = side_dict["left"]
            number_metabolites = len(
                [value for value in metabolite_dict.values() if value < 0]
            )
            if coefficient > 0:
                item = "products"
                SIDE = 1
                counter = side_dict["right"]
                number_metabolites = len(
                    [value for value in metabolite_dict.values() if value > 0]
                )
            # For metabolites: the labels must be 40 px higher in th y-axis
            # and the x-axis varies depending in the length (between 31-50)
            _up_or_down = cycle([-1, 1])
            if number_metabolites > 2:
                _up_or_down = cycle([-1])
            # Defining positions and labels depending on the Reaction-box and
            # the axis given
            if vertical:
                space_y = self.R_WIDTH / (number_metabolites + 1)
                dot_x = left_edge + counter * space_y
                dot_y = top_edge + (self.R_HEIGHT * 0.8 * SIDE) + 30
                label_x = dot_x + 10
                label_y = dot_y - 10 + SIDE * 40
            else:
                space_y = self.R_HEIGHT / (number_metabolites + 1)
                dot_x = left_edge + (self.R_WIDTH * 0.8 * SIDE) + 30
                dot_y = top_edge + counter * space_y
                label_x = dot_x - 30
                label_y = dot_y + 30 * next(_up_or_down) + 10
            # Check for shared metabolites only if metabolite is located in
            # reactant side
            if SIDE == 0:
                shared_node, old_reaction = self._find_shared(
                    metabolite=metabolite, products=shared_products
                )
            else:
                shared_node, old_reaction = str(), str()
            # If not empty, then defined new node
            if shared_node:
                node_number = shared_node
                debug_log.debug(
                    f'Metabolite "{metabolite}" in previous reaction '
                    f'"{old_reaction}" located in node "{shared_node}"'
                )
                # TODO: Move previous node
            else:
                # Add node metabolite-node to JsonDictionary. This has to be
                # omitted if previous reaction has a shared metabolite. Also,
                # change last variable to last node.
                self.add_metabolite(
                    x=dot_x,
                    y=dot_y,
                    label_x=label_x,
                    label_y=label_y,
                    bigg_id=metabolite,
                    # TODO: Add proper name
                    name=metabolite,
                    node_is_primary=False,
                )
                node_number = self._get_last_number(item="nodes") - 1
            # Increase number. TODO: find a better way
            counter += 1
            if coefficient < 0:
                side_dict["left"] = counter
            else:
                side_dict["right"] = counter
            # Add coefficient to reaction object
            reaction.add_metabolite(bigg_id=metabolite, coefficient=coefficient)
            # Add to node dictionary. Last minus one, since the node was
            # already added.
            # Either product or reactant
            self._overview[identifier][item].update({metabolite: coefficient})
            self._overview[identifier]["nodes"].update(
                {metabolite: node_number}
            )

    def add_segments(self, reaction: Reaction, metabolite_dict: dict):
        """
        Add the segments to given Reaction. This will make the visuals in
        Escher. The information about the nodes of metabolites in located in
        the JsonDictionary. A reaction will have 2 + number of metabolite as
        its number of segments.

        Args:
            metabolite_dict (dict): Dictionary with metabolites and their
                coefficients.
            reaction (Reaction): Reaction to extend.
        """
        # Defining identifier, the last number of segments and the markers
        identifier = reaction["bigg_id"]
        last = self._get_last_number(item="segments") - 1
        marker = {
            "first": self._overview[identifier]["nodes"]["_first"],
            "last": self._overview[identifier]["nodes"]["_last"],
            "middle": self._overview[identifier]["nodes"]["_middle"],
        }
        # From markers. They will be always 2.
        for node in ("first", "last"):
            last += 1
            reaction.add_segment(
                identifier=str(last),
                from_node_id=str(marker[node]),
                to_node_id=str(marker["middle"]),
            )
        for metabolite, coefficient in metabolite_dict.items():
            # Two due to the first two segments, and plus one as it represent
            # the actual Segment
            number = self._overview[identifier]["nodes"][metabolite]
            last += 1
            # Check whether reactant or product
            if coefficient < 0:
                reaction.add_segment(
                    identifier=str(last),
                    from_node_id=str(number),
                    to_node_id=str(marker["first"]),
                )
            elif coefficient > 0:
                reaction.add_segment(
                    identifier=str(last),
                    from_node_id=str(number),
                    to_node_id=str(marker["last"]),
                )
            else:
                # TODO: add exception
                raise Warning(f'Coefficient of "{metabolite}" cannot be 0')
        # Verify the number of Segments. They cannot have the same key of other
        # Segments from other reactions.

    def add_reaction(
        self,
        row: int,
        column: int,
        string: str,
        name: str,
        identifier: str,
        vertical: bool,
    ):
        """
        Parses and add given reaction string as a reaction for the
        JsonDictionary. It will automatically create all the necessary nodes
        and segments for the JSON.

        Args:
            string (str): Reaction string to be parsed.
            identifier (str): Identifier for the reaction.
            row (int): Row number from the visualization matrix.
            column (int): Column number of the visualization matrix.
            name (str): The name of the reaction.
            vertical (bool): If reaction should be displayed vertically.
        """
        # Add general data
        self._overview[identifier] = {
            "reactants": {},
            "products": {},
            "nodes": {},
            "position": Position(row=row, column=column),
            "index": "",
        }
        # Extract information for new reaction, nr of metabolites (string
        # representation)
        metabolite_dict = _convert_string(string=string)
        left_edge = self.R_WIDTH * column
        top_edge = self.R_HEIGHT * row
        # Labels
        if vertical:
            label_x = left_edge + (self.R_WIDTH / 2) + 40
            label_y = top_edge + (self.R_HEIGHT / 2)
        else:
            label_x = (left_edge + self.R_WIDTH / 2) - len(identifier) / 2 * 18
            label_y = top_edge + (self.R_HEIGHT) / 10
        # TODO: change this part
        reversibility = True
        reaction = Reaction(
            name=name,
            bigg_id=identifier,
            reversibility=reversibility,
            label_x=label_x,
            label_y=label_y,
            gene_reaction_rule="",
            genes=[],
            segments=dict(),
        )
        # Add nodes (metabolites and markers)
        self.map_metabolites(
            metabolite_dict=metabolite_dict,
            reaction=reaction,
            top_edge=top_edge,
            left_edge=left_edge,
            vertical=vertical,
        )
        self._add_reaction_markers(
            identifier=identifier,
            left_edge=left_edge,
            top_edge=top_edge,
            vertical=vertical,
        )
        # Add visual segments to reaction
        self.add_segments(reaction=reaction, metabolite_dict=metabolite_dict)
        # Define reaction number
        number = self._get_last_number(item="reactions")
        self.data["reactions"].update({str(number): reaction})
        self._overview[identifier]["index"] = str(number)
        debug_log.info(f'Reaction "{identifier}" added to the JsonDictionary.')

    def _reset(self):
        """
        Resets the data of JsonDictionary to its initial state. This helps
        when calling the function multiple times.
        """
        self._overview = {}
        self.data["reactions"] = {}
        self.data["nodes"] = {}
        self.reaction_scale = dict()

    def color_grading(
        self,
        color: List[Union[str, List[int], None]],
        min_max: List[float] = None,
        quantile: bool = False,
        max_steps: int = 100,
        n_steps: int = None,
    ):
        """
        Function that creates a color scale between two predefined colors. The
        number of color gradations corresponds to the number of fluxes.

        Parameters:
            color (list of str or list of list of int):
                list of two colors. The first color defines the endpoint for
                positive values, the second for negative values. The colors
                must be passed in their rgb representation.
            min_max(list of float,optional):
                List consisting of two floats. The first int describes the
                minimum, the second the maximum. These two values are
                additionally added as data values and at the same time ensure
                that all values greater or less than these are ignored when
                creating the gradient.
            quantile(bool, optional):
                Defines whether quantiles are to be used for the creation of
                the gradient. Otherwise, the steps are equally distributed
                between the minimum and maximum.
            n_steps (int, optional):
                Sets the number of color steps.
            max_steps(int, optional):
                Sets the maximum number of color steps.
        """

        # check if any flux values exist otherwise return
        if self.flux_solution is None:
            return

        # turn int arrays into numpy arrays
        color_positive = _color2np_rgb(color[0])
        color_negative = _color2np_rgb(color[1])
        color_intermediate = np.array([220, 220, 220], dtype=np.float32)

        try:
            flux = list(self.flux_solution.values())
        except AttributeError:
            flux = []

        # divide positive and negative values
        # bools used to handle the situation when both values of min_max are
        # positive or negative
        positive, negative, both_positive, both_negative = _divide_values(
            flux=flux, min_max=min_max
        )

        # array that will contain the configuration for escher
        reaction_scale = []

        # necessary variables for the loop that processes the positive
        # variables
        steps = len(positive)
        steps = min(steps, max_steps)

        # manually set steps overwrite the calculated steps
        if n_steps is not None:
            steps = math.floor(n_steps / 2)

        flux_values: np.ndarray

        try:
            if quantile:
                flux_values = np.linspace(0.0, 1.0, steps)

                # throws IndexError if positive == []
                flux_values = np.quantile(positive, flux_values)
            else:
                maximum = max(positive)
                start_value = maximum / steps

                # if both min_max values are positive the start is shifted
                # from zero to the set minimum
                if both_positive:
                    assert min_max is not None
                    start_value = min_max[0]

                flux_values = np.linspace(start_value, maximum, steps)

            with np.errstate(divide="ignore"):
                step_color = (color_intermediate - color_positive) / steps
            flux_color = color_intermediate.copy() - step_color

            flux_values.sort()
        except (ValueError, IndexError):
            flux_values = np.empty(shape=(0, 0))

        for flux_value in flux_values:
            reaction_scale.append(
                {
                    "type": "value",
                    "value": flux_value,
                    "color": "rgb(%d,%d,%d)"
                    % (flux_color[0], flux_color[1], flux_color[2]),
                }
            )

            flux_color -= step_color

        # add the intermediate step
        if not both_positive and not both_negative:
            reaction_scale.append(
                {"type": "value", "value": 0, "color": "rgb(220,220,220)"}
            )

        # updating of the variables for the negative portion of the flux values
        steps = len(negative)
        steps = min(steps, max_steps)

        # manually set steps overwrite the calculated steps
        if n_steps is not None:
            steps = math.floor(n_steps / 2)

        try:
            if quantile:
                flux_values = np.linspace(0.0, 1.0, steps)

                # throws IndexError if negative == []
                flux_values = np.quantile(negative, flux_values)
            else:
                minimum = min(negative)
                start_value = minimum / steps

                # if both min_max values are negative the start will be
                # shifted from zero to the set maximum
                if both_negative:
                    assert min_max is not None
                    start_value = min_max[1]
                flux_values = np.linspace(start_value, minimum, steps)

            with np.errstate(divide="ignore"):
                step_color = (color_intermediate - color_negative) / steps
            flux_color = color_intermediate - step_color

            flux_values.sort()
        except (ValueError, IndexError):
            flux_values = np.empty(shape=(0, 0))

        for flux_value in flux_values[::-1]:
            reaction_scale.append(
                {
                    "type": "value",
                    "value": flux_value,
                    "color": "rgb(%d,%d,%d)"
                    % (flux_color[0], flux_color[1], flux_color[2]),
                }
            )

            flux_color -= step_color

        self.reaction_scale = reaction_scale

    def visualize(
        self,
        filepath: Union[str, Path],
        vertical: bool = False,
        color: Optional[List[Union[str, List[int], None]]] = None,
        min_max: Optional[List[float]] = None,
        quantile: bool = False,
        max_steps: int = 100,
        n_steps: Optional[int] = None,
    ):
        """
        Saves the visualization of the JsonDictionary in given path as a HTML.
        Returns the builder for the JsonDictionary. If method is called in
        Jupyter or Qtconsole, it will show the embedded builder of the escher
        visualization. Else, it will open the default browser of the operating
        system and will load the previously saved HTML.

        .. note::
        Blank spaces are removed from the reactions.

        Args:
            filepath : Path, optional
                Path for the HTML. Defaults to "pathway.html" in the current
                working directory
            vertical : bool, optional
                Defines if pathway should be illustrated vertically.
                Defaults to False.
            color : list of str or list of lost of int, optional
                A list consisting of two entries. These define the endpoints
                for a color gradient. The entries can either be the colors as
                a list with three elements that define the RGB values or
                a string that defines the color name according to the css
                standard. If None is used here, no gradient will be generated.
                Defaults to None.
                Example: ["orange", "green"] or [[255, 165, 0], [0, 128, 0]]
            min_max: list of float, optional
                List consisting of two ints. The first int describes the
                minimum, the second the maximum. These two values are
                additionally added as data values and at the same time ensure
                that all values greater or less than these are ignored when
                creating the gradient.
                With Quantile = False, a data-independent gradient is created.
            quantile : bool, optional
                Defines whether quantiles are to be used for the creation of
                the gradient. Otherwise, the steps are equally distributed
                between the minimum and maximum.
            max_steps : int, optional
                Sets the maximum number of color steps.
            n_steps : int, optional
                Sets the maximum number of color steps.

        Returns:
            Builder: Escher builder object for the visualization

        See Also:
            Color names according to the css standard:
            https://www.w3schools.com/cssref/css_colors.asp
        """
        # Define path
        if isinstance(filepath, str):
            filepath = Path.cwd().joinpath(filepath)
        # Use relationship
        mapping = get_mapping(graph=self.graph)
        if vertical:
            mapping = transpose(matrix=mapping)
            # Modify Reaction-Box
            self.R_HEIGHT, self.R_WIDTH = self.R_WIDTH, self.R_HEIGHT
        # Modify canvas
        self.CANVAS_HEIGHT = self.R_HEIGHT * len(mapping)
        self.CANVAS_WIDTH = self.R_WIDTH * len(mapping[0])
        # Use reaction information
        for index_j, row in enumerate(mapping):
            for index_i, reaction in enumerate(row):
                # Add reactions only not 0
                if reaction == 0:
                    continue
                self.add_reaction(
                    row=index_j,
                    column=index_i,
                    name=reaction,
                    string=self.reaction_strings[reaction],
                    identifier=reaction,
                    vertical=vertical,
                )

        if color is not None:
            self.color_grading(
                color=color,
                min_max=min_max,
                quantile=quantile,
                max_steps=max_steps,
                n_steps=n_steps,
            )

        builder = escher.Builder(
            # Check how reaction_styles behaves
            reaction_styles=["color", "text"],
            map_name=self.data["head"]["map_name"],
            map_json=self.json_dump(),
            reaction_scale=self.reaction_scale,
        )
        # This statement is needed, otherwise, all reactions labels will
        # appear with "(nd)".
        if self.flux_solution:
            builder.reaction_data = self.flux_solution
        builder.save_html(filepath=filepath)

        # FIXME: temporal solution to https://github.com/Toepfer-Lab/escher-legacy/pull/6
        f = fileinput.FileInput(filepath, inplace=True)
        for line in f:
            if f.lineno() == 6:
                print(line.replace("1.7.4", "1.7.3"), end="")
            else:
                print(line, end="")
        f.close()

        debug_log.info(f'Visualization saved in "{filepath}"')

        # Cleaning Up
        self._reset()
        return builder



"""
Debugging configuration for sub-package visualization.

Configures the debug logging tool. The format follows the syntax: `(asctime)
(levelname) (message)`. The logs are saved in the logs directory and includes
'visualization' in the file and its date. The default level is set to DEBUG.

The name of logger variable is `visualization`
"""
import datetime as dt
import logging
from pathlib import Path


debug_log = logging.getLogger("visualization")
debug_log.setLevel(logging.DEBUG)

FORMAT_STR = "[%(asctime)s] %(levelname)s %(message)s"
TIME_STR = "%H:%M:%S"

# Format
debug_formatter = logging.Formatter(FORMAT_STR, TIME_STR)

log_dir = Path("logs").absolute()
log_dir.mkdir(exist_ok=True)

log_path = log_dir.joinpath(
    f"visualization_{dt.date.today().strftime('%Y%m%d')}.log"
)
debug_handler = logging.FileHandler(log_path, mode="a+")
debug_handler.setFormatter(debug_formatter)

debug_log.addHandler(debug_handler)
